# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import asyncio

from x2py.deserializer import Deserializer
from x2py.serializer import Serializer
from x2py.util.trace import Trace
from x2py.links.link_session import LinkSession

class TcpSession(LinkSession, asyncio.Protocol):
    def __init__(self, link):
        LinkSession.__init__(self, link)
        asyncio.Protocol.__init__(self)
        self.transport = None

    def cleanup(self):
        super(TcpSession, self).cleanup()
        self.transport.close()

    def connected(self):
        if self.transport is not None:
            if self.transport._sock is not None:
                return (self.transport._sock.fileno() >= 0)
        return False

    def connection_made(self, transport):
        self.transport = transport
        self.link.init_session(self)

    def connection_lost(self, exc=None):
        self.link.on_disconnect(self.handle, self)
        self.transport = None
        super(TcpSession, self).connection_lost(exc)

    def data_received(self, data):
        if not data:
            self.connection_lost(self.transport)
            return
        Trace.trace("{} received {}", self.link.name, data)
        self.on_receive(data)

    def build_header(self, buffer, transformed):
        length = len(buffer)
        header = 1 if transformed else 0
        header = header | (length << 1)

        result = bytearray()
        Serializer.write_variable(result, header)
        return result

    def parse_header(self, deserializer):
        pos = deserializer.pos
        try:
            header, num_bytes = deserializer.read_variable32()
        except:
            return 0, 0, False
        finally:
            deserializer.pos = pos

        length = header >> 1
        transformed = ((header & 1) != 0)
        return num_bytes, length, transformed

    def _send(self, data):
        Trace.trace("{} sending {}", self.link.name, data)
        self.transport.write(data)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import asyncore
import socket
from threading import Thread

from x2py.util.trace import Trace
from x2py.links.server_link import ServerLink
from x2py.links.asyncore.tcp_session import TcpSession

class TcpServer(ServerLink):
    """TCP/IP server link based on the asyncore module."""

    class Dispatcher(asyncore.dispatcher):
        def __init__(self, owner):
            asyncore.dispatcher.__init__(self, map=owner.map)
            self.owner = owner
        def handle_accept(self):
            self.owner.handle_accept()

    def __init__(self, name):
        super(TcpServer, self).__init__(name)
        self.map = {}
        self.dispatcher = TcpServer.Dispatcher(self)
        self.thread = Thread(target=self._loop)

    def cleanup(self):
        asyncore.close_all(map=self.map)
        self.dispatcher.close()

        self.thread.join()

        super(TcpServer, self).cleanup()

    def listen(self, host, port):
        self.dispatcher.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.dispatcher.bind((host, port))
        self.dispatcher.listen(5)

        Trace.info("listening on {}:{}", host, port)

        self.thread.start()

    def handle_accept(self):
        pair = self.dispatcher.accept()
        if pair is not None:
            sock, addr = pair
            session = TcpSession(self, sock)
            session.connection_made()

    def _loop(self):
        asyncore.loop(map=self.map)

    def _on_connect(self, result, context):
        super(TcpServer, self)._on_connect(result, context)
        if result:
            peername = context.socket.getpeername()
            Trace.info("accepted from {}:{}", peername[0], peername[1])
            context.peername = peername

    def _on_disconnect(self, handle, context):
        super(TcpServer, self)._on_disconnect(handle, context)
        peername = context.peername
        Trace.info("disconnected from {}:{}", peername[0], peername[1])



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.yields.wait_for_completion import WaitForCompletion
from x2py.yields.wait_for_event import WaitForEvent, WaitForResponse
from x2py.yields.wait_for_next import WaitForNext
from x2py.yields.wait_for_seconds import WaitForSeconds



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from threading import Lock

from x2py.event import Event
from x2py.event_queue import EventQueue
from x2py.flow import Flow

class EventBasedFlow(Flow):
    """Abstract base class for event-based (waiting) execution flows."""

    def __init__(self, name=None):
        super(EventBasedFlow, self).__init__(name)
        self.queue = EventQueue()
        self._lock = Lock()

    def feed(self, event):
        if event is None or not isinstance(event, Event):
            raise TypeError()
        self.queue.enqueue(event)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

class Formatter(object):
    def desc(self):
        raise NotImplementedError()

    def format(self, unit, out_dir):
        raise NotImplementedError()

    def is_up_to_date(self, path, out_dir):
        raise NotImplementedError()

class FormatterContext(object):
    def format_cell(self, definition):
        raise NotImplementedError()

    def format_consts(self, definition):
        raise NotImplementedError()

    def format_reference(self, reference):
        raise NotImplementedError()



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from array import array
from copy import copy

from x2py.deserializer import Deserializer
from x2py.serializer import Serializer
from x2py.util.hash import HASH_SEED, hash_update

class Fingerprint(object):
    """Manages a fixed-length compact array of bit values.
        (zero-based indexing) """

    def __init__(self, arg):
        if isinstance(arg, int):
            self._ctor(arg)
        elif isinstance(arg, Fingerprint):
            self._copy_ctor(arg)
        else:
            raise TypeError()

    def _ctor(self, length):
        if length < 0:
            raise ValueError()

        self.length = length
        self.block = 0
        self.blocks = None

        if length > 32:
            length -= 32
            self.blocks = array('L', [0] * (((length - 1) >> 5) + 1))

    def _copy_ctor(self, other):
        self.length = other.length
        self.block = other.block
        if other.blocks is not None:
            self.blocks = copy(other.blocks)
        else:
            self.blocks = None

    @property
    def length_in_bytes(self):
        return ((self.length - 1) >> 3) + 1

    def get(self, index):
        """Gets the bit value at the specified index."""
        if index < 0:
            raise ValueError()
        # Doesn't throw on over-indexing
        if index >= self.length:
            return False

        if index < 32:
            return ((self.block & (1 << (index & 0x1f))) != 0)
        else:
            index -= 32
            return ((self.blocks[index >> 5] & (1 << (index & 0x1f))) != 0)

    def touch(self, index):
        """Sets the bit value at the specified index."""
        if index < 0:
            raise ValueError()
        # Allow over-indexing, returning False by default.
        if self.length <= index:
            return False

        if index < 32:
            self.block |= (1 << (index & 0x1f))
        else:
            index -= 32
            self.blocks[index >> 5] |= (1 << (index & 0x1f))

    def wipe(self, index):
        """Clears the bit value at the specified index."""
        if index < 0 or self.length <= index:
            raise ValueError()

        if index < 32:
            self.block &= ~(1 << (index & 0x1f))
        else:
            index -= 32
            self.blocks[index >> 5] &= ~(1 << (index & 0x1f))

    def deserialize(self, deserializer):
        length, _ = deserializer.read_nonnegative()
        num_bytes = ((length - 1) >> 3) + 1
        num_blocks = ((num_bytes - 1) >> 2) + 1

        i = count = 0
        block = 0
        while i < 4 and count < num_bytes:
            b = deserializer.read_byte(None)
            if count < self.length_in_bytes:
                block = block | (b << (i << 3))
            i += 1
            count += 1
        self.block = block

        j = 0
        while j < num_blocks:
            i = 0
            block = 0
            while i < 4 and count < num_bytes:
                b = deserializer.read_byte(None)
                if count < self.length_in_bytes:
                    block = block | (b << (i << 3))
                i += 1
                count += 1
            if self.blocks is not None and j < len(self.blocks):
                self.blocks[j] = block
            j += 1

    def get_length(self):
        return Serializer.get_length_nonnegative(self.length) + self.length_in_bytes

    def serialize(self, serializer):
        serializer.write_nonnegative(self.length)

        i = count = 0
        while i < 4 and count < self.length_in_bytes:
            serializer.write_byte(None, (self.block >> (i << 3)) & 0x0ff)
            i += 1
            count += 1

        if self.blocks is None:
            return

        for block in self.blocks:
            i = 0
            while i < 4 and count < self.length_in_bytes:
                serializer.write_byte(None, (block >> (i << 3)) & 0x0ff)
                i += 1
                count += 1

    def equivalent(self, other):
        if self is other:
            return True
        if not isinstance(other, Fingerprint) or self.length < other.length:
            return False
        if (self.block & other.block) != other.block:
            return False
        if other.blocks is not None:
            count = len(other.blocks)
            i = 0
            while i < count:
                block = other.blocks[i]
                if (self.blocks[i] & block) != block:
                    return False
                i += 1
        return True

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Fingerprint) or self.length != other.length:
            return False
        if self.block != other.block:
            return False
        if self.blocks is not None:
            count = len(self.blocks)
            i = 0
            while i < count:
                if self.blocks[i] != other.blocks[i]:
                    return False
                i += 1
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        value = hash_update(HASH_SEED, self.length)
        value = hash_update(value, self.block)
        if self.blocks is not None:
            for block in self.blocks:
                value = hash_update(value, block)
        return value

    def __lt__(self, other):
        if self is other:
            return False
        if self.length < other.length:
            return True
        if self.length > other.length:
            return False
        # assert self.length == other.length
        if self.blocks is not None:
            i = len(self.blocks) - 1
            while i >= 0:
                block = self.blocks[i]
                other_block = other.blocks[i]
                if block < other_block:
                    return True
                if block > other_block:
                    return False
                i -= 1
        if self.block >= other.block:
            return False
        return True



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from threading import Lock

from x2py.case import Case
from x2py.event_factory import EventFactory

class Link(Case):
    """Common abstract base class for link cases."""

    names = set()
    _lock = Lock()

    def __init__(self, name):
        super(Link, self).__init__()
        self.event_factory = EventFactory.new()
        with Link._lock:
            if name in Link.names:
                raise ValueError("link name '{}' is already in use".format(name))
            self._name = name
            Link.names.add(name)

    @property
    def name(self):
        return self._name

    def cleanup(self):
        with Link._lock:
            if self.name in Link.names:
                Link.names.remove(self.name)

    def close(self):
        self.cleanup()

    def create_event(self, type_id):
        # Try link-local event factory first.
        result = self.event_factory.create(type_id)
        if result:
            return result
        # If not fount, try the global event factory.
        return EventFactory.create(type_id)

    def send(self, event):
        raise NotImplementedError()

    def _teardown(self):
        self.close()
        super(Link, self)._teardown()



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.event_sink import EventSink
from x2py.util.trace import Trace

class Case(EventSink):
    """Represents a set of application logic."""

    def __init__(self):
        super(Case, self).__init__()

    def setup_with(self, flow):
        """Initializes this case with the specified holding flow."""
        self.flow = flow

        from x2py.flow import Flow
        backup = Flow.thread_local.current
        Flow.thread_local.current = flow

        self._setup()

        Flow.thread_local.current = backup

    def teardown_with(self, flow):
        """Cleans up this case with the specified holding flow."""
        from x2py.flow import Flow
        backup = Flow.thread_local.current
        Flow.thread_local.current = flow

        self._teardown()

        Flow.thread_local.current = backup

        self.cleanup()  # eventsink cleanup

    def setup(self):
        """Overridden by subclasses to build a initialization chain."""
        pass

    def teardown(self):
        """Overridden by subclasses to build a cleanup chain."""
        pass

    def on_start(self):
        """Overridden by subclasses to build a flow startup handler chain."""
        pass

    def on_stop(self):
        """Overridden by subclasses to build a flow shutdown handler chain."""
        pass

    def _setup(self):
        """Called internally when this case is initialized."""
        self.setup()

    def _teardown(self):
        """Called internally when this case is cleaned up."""
        self.teardown()

class CaseStack(object):
    """Handles a group of cases."""

    def __init__(self):
        self.cases = []
        self.activated = False

    def add(self, case):
        if case is None or not isinstance(case, Case):
            raise TypeError()
        if case in self.cases:
            return False
        self.cases.append(case)
        return True

    def remove(self, case):
        if case is None or not isinstance(case, Case):
            raise TypeError()
        if case not in self.cases:
            return False
        self.cases.remove(case)
        return True

    def setup_with(self, flow):
        if self.activated:
            return
        self.activated = True
        for case in self.cases:
            Trace.trace("casestack: setting up case {}", type(case).__name__)
            case.setup_with(flow)

    def teardown_with(self, flow):
        if not self.activated:
            return
        self.activated = False
        for case in reversed(self.cases):
            try:
                Trace.trace("casestack: tearing down case {}", type(case).__name__)
                case.teardown_with(flow)
            except BaseException as ex:
                Trace.error("{} {} teardown: {}", flow.name, type(case).__name__, ex)

    def on_start(self):
        for case in self.cases:
            case.on_start()

    def on_stop(self):
        for case in reversed(self.cases):
            try:
                case.on_stop()
            except:
                pass



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from threading import Thread

from x2py.builtin_events import *
from x2py.event import EventProxy
from x2py.flow import Flow
from x2py.flows.event_based_flow import EventBasedFlow
from x2py.util.trace import Trace

class MultiThreadFlow(EventBasedFlow):
    def __init__(self, name=None, num_threads=2):
        super(MultiThreadFlow, self).__init__(name)
        self.threads = []
        self.num_threads = num_threads

    def start(self):
        with self._lock:
            if len(self.threads) != 0:
                return

            self._setup()
            self.cases.setup_with(self)

            for i in range(self.num_threads):
                thread = Thread(target=self)
                thread.setName("{} {}".format(self.name, i + 1))
                thread.start()
                self.threads.append(thread)

            self.queue.enqueue(FlowStart())

    def stop(self):
        with self._lock:
            if len(self.threads) == 0:
                return

            self.queue.close(FlowStop())

            for thread in self.threads:
                thread.join()
            del self.threads[:]

            self.cases.teardown_with(self)
            self._teardown()

    def __call__(self):
        Flow.thread_local.current = self
        Flow.thread_local.event_proxy = EventProxy()
        Flow.thread_local.handler_chain = []

        while True:
            event = self.queue.dequeue()
            if event is None:
                break
            self.dispatch(event)

        Flow.thread_local.handler_chain = None
        Flow.thread_local.event_proxy = None
        Flow.thread_local.current = None




# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.buffer_transform import BufferTransform

class Inverse(BufferTransform):
    """A trivial example of BufferTransform that just invert every bit."""

    def cleanup(self):
        pass

    def clone(self):
        return Inverse()

    def handshake_block_length(self):
        return 0

    def init_handshake(self):
        return None
    def handshake(self, challenge):
        return None
    def fini_handshake(self, response):
        return True

    def transform(self, buffer):
        result = bytearray()
        for b in buffer:
            result.append(~b & 0x0ff)
        return result
    def inverse_transform(self, buffer):
        result = bytearray()
        for b in buffer:
            result.append(~b & 0x0ff)
        return result



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.util.ranged_int_pool import RangedIntPool

class WaitHandlePool(object):
    pool = RangedIntPool(1, 65536, True)

    def __init__(self):
        raise AssertionError()

    @staticmethod
    def acquire():
        return WaitHandlePool.pool.acquire()

    @staticmethod
    def release(handle):
        return WaitHandlePool.pool.release(handle)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from datetime import timedelta

from x2py.builtin_events import TimeoutEvent
from x2py.flow import Flow
from x2py.flows.time_flow import TimeFlow

class WaitForNext(object):
    def __init__(self, coroutine, result):
        self.coroutine = coroutine
        self.result = result
        e = TimeoutEvent().setattrs(key=self)
        self.token = Flow.bind(e, self.on_event)
        e.post()

    def on_event(self, e):
        Flow.unbind_with(self.token)
        self.coroutine.result = self.result
        self.coroutine.next()



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

HASH_SEED = 17

def hash_update(seed, value):
    return int(((seed << 5) + seed) ^ value)

class Hash(object):
    def __init__(self, code=HASH_SEED):
        self.code = code

    def update(self, value):
        self.code = hash_update(self.code, value)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from __future__ import print_function

import os
import sys

from xpiler import __version__
from xpiler.options import Options

def _init_handlers():
    result = {}
    from xpiler.xml_handler import XmlHandler
    result['.xml'] = XmlHandler()
    return result

def _init_formatters():
    result = {}
    from xpiler.python3_formatter import Python3Formatter
    result['py'] = Python3Formatter()
    return result

class Main(object):
    handlers = _init_handlers()
    formatters = _init_formatters()
    options = Options()

    def __init__(self):
        self.formatter = Main.formatters[Main.options.spec]
        self.subdirs = []
        self.error = False

    @staticmethod
    def run(argv):
        print("x2py.xpiler v{}".format(__version__))
        args = Main.options.parse(argv[1:])
        if len(args) < 1:
            print("error: at least one input path is required", file=sys.stderr)
            sys.exit(2)

        main = Main()
        for path in args:
            main.process(path)
        return 1 if main.error else 0

    def process(self, path):
        if os.path.isdir(path):
            self._process_dir(path)
        elif os.path.isfile(path):
            self._process_file(path)
        else:
            print(path + " doesn't exist.", file=sys.stderr)
            self.error = True

    def _process_dir(self, path):
        print("Directory " + os.path.abspath(path))
        for entry in os.listdir(path):
            entry_path = os.path.join(path, entry)
            if os.path.isdir(entry_path):
                if Main.options.recursive:
                    self.subdirs.append(entry)
                    self._process_dir(entry_path)
                    self.subdirs.pop()
            else:
                self._process_file(entry_path)

    def _process_file(self, path):
        filename = os.path.basename(path)
        basename, extension = os.path.splitext(filename)

        if Main.options.out_dir is None:
            out_dir = os.path.dirname(path)
        else:
            out_dir = os.path.join(Main.options.out_dir, os.path.sep.join(self.subdirs))

        if (extension.lower() not in Main.handlers) or (not Main.options.forced
                and self.formatter.is_up_to_date(path, out_dir)):
            return

        print(filename)

        handler = Main.handlers[extension.lower()]
        result, unit = handler.handle(path)
        if result == False:
            self.error = True
        if self.error or unit is None:
            return

        unit.basename = basename

        if (out_dir and len(out_dir) != 0) and not os.path.exists(out_dir):
            os.makedirs(out_dir)

        result = self.formatter.format(unit, out_dir)
        if result == False:
            self.error = True



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from datetime import datetime, timedelta
import heapq

from x2py.builtin_events import *
from x2py.event import EventProxy
from x2py.flow import Flow
from x2py.flows.frame_based_flow import FrameBasedFlow
from x2py.util.rwlock import ReadLock, WriteLock, ReadWriteLock
from x2py.util.trace import Trace

class Timer(object):
    class Tag(object):
        def __init__(self, time_delta, next_utc_time=datetime.utcnow()):
            self.time_delta = time_delta
            self.next_utc_time = next_utc_time

    class Repeater(object):
        def __init__(self, owner):
            self.rwlock = ReadWriteLock()
            self.map = {}
            self.owner = owner
            self.default_case = None

        def add(self, state, time_tag):
            with self.rwlock.wlock():
                if state is not None:
                    self.map[state] = time_tag
                else:
                    default_case = time_tag

        def remove(self, state):
            with self.rwlock.wlock():
                if state is not None:
                    del self.map[state]
                else:
                    default_case = None

        def tick(self, utc_now):
            with self.rwlock.rlock():
                if self.default_case is not None:
                    self.try_fire(utc_now, None, self.default_case)
                if self.map:
                    for key, value in self.map.items():
                        self.try_fire(utc_now, key, value)

        def try_fire(self, utc_now, state, time_tag):
            if utc_now > time_tag.next_utc_time:
                self.owner.callback(state)
                time_tag.next_utc_time = utc_now + time_tag.time_delta

    def __init__(self, callback):
        self.rwlock = ReadWriteLock()
        self.reserved = []
        self.repeater = Timer.Repeater(self)
        self.callback = callback

    def reserve(self, state, time_delta):
        return self.reserve_at_utc_time(state, datetime.utcnow() + time_delta)

    def reserve_at_utc_time(self, state, utc_time):
        with self.rwlock.wlock():
            heapq.heappush(self.reserved, (utc_time, state))
        return (utc_time, state)

    def cancel(self, timer_token):
        with self.rwlock.wlock():
            self.reserved.remove(timer_token)
            heapq.heapify(self.reserved)

    def reserve_repetition(self, state, time_delta, next_utc_time=None):
        if next_utc_time is None:
            next_utc_time = datetime.utcnow() + time_delta
        self.repeater.add(state, Timer.Tag(time_delta, next_utc_time))

    def cancel_repetition(self, state):
        self.repeater.remove(state)

    def tick(self):
        utc_now = datetime.utcnow()
        events = None
        with self.rwlock.rlock() as rlock:
            if self.reserved:
                peeked = self.reserved[0]
                if utc_now >= peeked[0]:
                    events = []
                    rlock.upgrade()
                    while utc_now >= peeked[0]:
                        popped = heapq.heappop(self.reserved)
                        events.append(popped[1])
                        if not self.reserved:
                            break
                        peeked = self.reserved[0]

        if events:
            for event in events:
                self.callback(event)

        self.repeater.tick(utc_now)


class TimeFlow(FrameBasedFlow):
    rwlock = ReadWriteLock()
    instances = {}

    def __init__(self, name):
        super(TimeFlow, self).__init__(name=name)
        self.timer = Timer(self.on_timer)

    @staticmethod
    def get(name='default'):
        with TimeFlow.rwlock.rlock():
            flow = None
            if name not in TimeFlow.instances:
                flow_name = "TimeFlow.{}".format(name)
                flow = TimeFlow(flow_name)
                TimeFlow.instances[name] = flow
                flow.start()
                flow.attach()
            else:
                flow = TimeFlow.instances[name]
        return flow

    def reserve(self, event, time_delta):
        return self.timer.reserve(event, time_delta)

    def reserve_at_utc_time(self, event, utc_time):
        return self.timer.reserve_at_utc_time(event, utc_time)

    def cancel(self, timer_token):
        self.timer.cancel(timer_token)

    def reserve_repetition(self, event, time_delta, next_utc_time=None):
        if next_utc_time is None:
            next_utc_time = datetime.utcnow() + time_delta
        self.timer.reserve_repetition(event, time_delta, next_utc_time)

    def cancel_repetition(self, event):
        self.timer.cancel_repetition(event)

    def update(self):
        self.timer.tick()

    def on_timer(self, event):
        event.post()



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

class TraceLevel(object):
    ALL = 0
    TRACE = 1
    DEBUG = 2
    INFO = 3
    WARNING = 4
    ERROR = 5
    NONE = 6

    rmap = { 0: 'ALL', 1: 'TRACE', 2: 'DEBUG', 3: 'INFO', 4: 'WARNING', 5: 'ERROR', 6: 'NONE' }

    @staticmethod
    def name(value):
        return TraceLevel.rmap[value]

class Trace(object):
    """Represents the tracing helper class."""

    handler = None
    level = TraceLevel.INFO

    def __init__(self):
        raise AssertionError()

    @staticmethod
    def emit(level, message):
        if Trace.handler is None or Trace.level > level:
            return
        Trace.handler(level, message)
    @staticmethod
    def emit(level, format_string, *args, **kwargs):
        if Trace.handler is None or Trace.level > level:
            return
        Trace.handler(level, format_string.format(*args, **kwargs))

    @staticmethod
    def trace(message):
        Trace.emit(TraceLevel.TRACE, message)
    @staticmethod
    def trace(format_string, *args, **kwargs):
        Trace.emit(TraceLevel.TRACE, format_string, *args, **kwargs)

    @staticmethod
    def debug(message):
        Trace.emit(TraceLevel.DEBUG, message)
    @staticmethod
    def debug(format_string, *args, **kwargs):
        Trace.emit(TraceLevel.DEBUG, format_string, *args, **kwargs)

    @staticmethod
    def info(message):
        Trace.emit(TraceLevel.INFO, message)
    @staticmethod
    def info(format_string, *args, **kwargs):
        Trace.emit(TraceLevel.INFO, format_string, *args, **kwargs)

    @staticmethod
    def warn(message):
        Trace.emit(TraceLevel.WARNING, message)
    @staticmethod
    def warn(format_string, *args, **kwargs):
        Trace.emit(TraceLevel.WARNING, format_string, *args, **kwargs)

    @staticmethod
    def error(message):
        Trace.emit(TraceLevel.ERROR, message)
    @staticmethod
    def error(format_string, *args, **kwargs):
        Trace.emit(TraceLevel.ERROR, format_string, *args, **kwargs)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from bisect import bisect

from x2py.event import Event, EventProxy
from x2py.fingerprint import Fingerprint
from x2py.util.atomic import AtomicInt
from x2py.util.rwlock import ReadLock, WriteLock, ReadWriteLock

def binary_search(a, x):
    index = bisect(a, x)
    if index and (a[index - 1] == x):
        return index - 1  # right-most index of the found
    else:
        return ~index  # insertion point

class Slot(Fingerprint):
    """Extends Fingerprint to hold an additional reference count."""

    def __init__(self, fingerprint):
        super(Slot, self).__init__(fingerprint)
        self.ref_count = AtomicInt(1)

    def add_ref(self):
        self.ref_count.increment()

    def remove_ref(self):
        return self.ref_count.decrement()

class Binding(object):
    """Manages evnet-handler bindings."""

    class _Filter(object):
        def __init__(self):
            self.map = {}

        def add(self, type_id, fingerprint):
            slots = self.map.get(type_id)
            if slots is None:
                slots = []
                self.map[type_id] = slots
            slot = Slot(fingerprint)
            index = binary_search(slots, slot)
            if index >= 0:
                slots[index].add_ref()
            else:
                index = ~index
                slots.insert(index, slot)

        def get(self, type_id):
            return self.map.get(type_id)

        def remove(self, type_id, fingerprint):
            slots = self.map.get(type_id)
            if slots is None:
                return
            index = binary_search(slots, Slot(fingerprint))
            if index < 0:
                return
            if slots[index].remove_ref() == 0:
                del slots[index]
            if len(slots) == 0:
                del self.map[type_id]

    def __init__(self):
        self.map = {}  # event => list of handlers
        self.filter = Binding._Filter()
        self.rwlock = ReadWriteLock()

    def bind(self, event, handler):
        with self.rwlock.wlock():
            handlers = self.map.get(event)
            if handlers is None:
                handlers = []
                self.map[event] = handlers

            token = (event, handler)
            if handler not in handlers:
                handlers.append(handler)
                self.filter.add(event.type_id(), event.fingerprint)

                if hasattr(handler, '__self__'):
                    from x2py.event_sink import EventSink
                    target = handler.__self__
                    if isinstance(target, EventSink):
                        target._add_binding(token)
            return token

    def _unbind_(self, event, handler):
        handlers = self.map.get(event)
        if handler is None:
            return
        if handler not in handlers:
            return
        handlers.remove(handler)
        if len(handlers) == 0:
            del self.map[event]
        self.filter.remove(event.type_id(), event.fingerprint)

    def _unbind(self, event, handler):
        with self.rwlock.wlock():
            self._unbind_(event, handler)

    def unbind(self, event, handler):
        with self.rwlock.wlock():
            self._unbind_(event, handler)

            if hasattr(handler, '__self__'):
                from x2py.event_sink import EventSink
                target = handler.__self__
                if isinstance(target, EventSink):
                    target._remove_binding((event, handler))

    def build_handler_chain(self, event, event_proxy, handler_chain):
        event_proxy.event = event
        tag = event.type_tag()
        fingerprint = event.fingerprint
        with self.rwlock.rlock():
            while tag is not None:
                type_id = tag.type_id
                event_proxy.type_id = type_id
                slots = self.filter.get(type_id)
                if slots is not None:
                    for slot in slots:
                        if fingerprint.equivalent(slot):
                            event_proxy.fingerprint = slot

                            handlers = self.map.get(event_proxy)
                            if handlers is not None:
                                handler_chain += handlers
                tag = tag.base




# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

class Config(object):
    heartbeat_interval = 5  # in seconds

    class Coroutine(object):
        default_timeout = 60 # in seconds


# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from threading import Thread

from x2py.builtin_events import *
from x2py.event import EventProxy
from x2py.flow import Flow
from x2py.flows.event_based_flow import EventBasedFlow
from x2py.util.trace import Trace

class SingleThreadFlow(EventBasedFlow):
    def __init__(self, name=None):
        super(SingleThreadFlow, self).__init__(name)
        self.thread = None

    def start(self):
        with self._lock:
            if self.thread is not None:
                return

            self._setup()
            self.cases.setup_with(self)

            self.thread = Thread(target=self)
            self.thread.setName(self.name)
            self.thread.start()

            self.queue.enqueue(FlowStart())

    def stop(self):
        with self._lock:
            if self.thread is None:
                return

            self.queue.close(FlowStop())

            self.thread.join()
            self.thread = None

            self.cases.teardown_with(self)
            self._teardown()

    def __call__(self):
        Flow.thread_local.current = self
        Flow.thread_local.event_proxy = EventProxy()
        Flow.thread_local.handler_chain = []

        while True:
            event = self.queue.dequeue()
            if event is None:
                break
            self.dispatch(event)

        Flow.thread_local.handler_chain = None
        Flow.thread_local.event_proxy = None
        Flow.thread_local.current = None




# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

class LinkStrategy(object):
    """Common abstract base class for link strategies."""

    def __init__(self):
        self.link = None

    def setup(self):
        pass
    def teardown(self):
        pass

class SessionBasedLinkStrategy(LinkStrategy):
    """Common abstract base class for session-based link strategies."""

    def before_session_setup(self, link_session):
        pass
    def after_session_teardown(self, link_session):
        pass

class LinkSessionStrategy(object):
    """Common abstract base class for link session strategies."""

    def __init__(self):
        self.session = None

    def process(self, event):
        return False

class ChannelStrategy(SessionBasedLinkStrategy):
    """Common abstract base class for communication channel strategies."""

    class SubStrategy(LinkSessionStrategy):
        def release(self):
            pass
        def before_send(self, buffer):
            return False
        def after_receive(self, buffer):
            pass

    def init_handshake(self, link_session):
        pass
    def cleanup(self):
        pass

class HeartbeatStrategy(SessionBasedLinkStrategy):
    """Common abstract base class for heartbeat strategies."""

    class SubStrategy(LinkSessionStrategy):
        def __init__(self):
            super(HeartbeatStrategy.SubStrategy, self).__init__()
            self.marked = False

        def on_heartbeat(self):
            return False
        def on_receive(self):
            pass
        def on_send(self, event):
            pass

    def on_heartbeat(self):
        pass



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

class Unit(object):
    """Represents a single definition unit."""

    def __init__(self):
        self.basename = ""
        self.namespace = ""
        self.references = []
        self.definitions = []
        self.is_builtin = False

class Reference(object):
    def __init__(self, target=""):
        self.target = target

    def format(self, context):
        context.format_reference(self)

class Definition(object):
    def __init__(self, name=""):
        self.name = name

    def format(self, context):
        raise NotImplementedError()

class Constant(object):
    def __init__(self, name="", value=""):
        self.name = name
        self.value = value

class Consts(Definition):
    def __init__(self, name, type):
        super(Consts, self).__init__(name)
        self.type = type
        self.constants = []

    def format(self, context):
        context.format_consts(self)

class Property(object):
    def __init__(self, name="", default_value=""):
        self.index = 0
        self.name = name
        self.typespec = None
        self.default_value = default_value

class Cell(Definition):
    def __init__(self):
        super(Cell, self).__init__()
        self.base = ""
        self.base_class = ""
        self.is_event = False
        self.local = False
        self.properties = []

    def has_properties(self):
        return (len(self.properties) != 0)

    def format(self, context):
        context.format_cell(self)

class Event(Cell):
    def __init__(self):
        super(Event, self).__init__()
        self.id = ""
        self.is_event = True


# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

class TypeSpec(object):
    def __init__(self, typestr, details):
        self.typestr = typestr
        self.details = details  # list(TypeSpec)

    def metaprop(self, name):
        tokens = []
        tokens.append('MetaProperty(')
        if name is not None:
            tokens.append("'{}'".format(name))
        else:
            tokens.append('None')
        tokens.append(", {}".format(Types.get_type_index(self.typestr)))
        if not Types.is_builtin(self.typestr):
            tokens.append(", runtime_type={}".format(self.typestr))
        if self.details is not None and len(self.details) != 0:
            tokens.append(', details=[ ')
            for index, detail in enumerate(self.details):
                if index:
                    tokens.append(', ')
                tokens.append(detail.metaprop(None))
            tokens.append(' ]')
        tokens.append(')')
        return ''.join(tokens)

    def __str__(self):
        tokens = [ self.type ]
        if self.details is not None and len(self.details) != 0:
            tokens.append('(')
            for index, detail in enumerate(self.details):
                if index:
                    tokens.append(', ')
                tokens.append(str(detail))
            tokens.append(')')
        return ''.join(tokens)

class TypeProperty(object):
    def __init__(self, is_primitive, is_collective, detail_required, index):
        self.is_primitive = is_primitive
        self.is_collective = is_collective
        self.detail_required = detail_required
        self.index = index

def _init_types():
    result = {}
    # Primitive types
    result["bool"] = TypeProperty(True, False, False, 1)
    result["byte"] = TypeProperty(True, False, False, 2)
    result["int8"] = TypeProperty(True, False, False, 3)
    result["int16"] = TypeProperty(True, False, False, 4)
    result["int32"] = TypeProperty(True, False, False, 5)
    result["int64"] = TypeProperty(True, False, False, 6)
    result["float32"] = TypeProperty(True, False, False, 7)
    result["float64"] = TypeProperty(True, False, False, 8)
    result["string"] = TypeProperty(True, False, False, 9)
    result["datetime"] = TypeProperty(True, False, False, 10)
    # Collective types
    result["bytes"] = TypeProperty(False, True, False, 11)
    result["list"] = TypeProperty(False, True, True, 13)
    result["map"] = TypeProperty(False, True, True, 14)
    # Non-serializable
    result["object"] = TypeProperty(False, False, False, 15)
    return result

class Types(object):
    map = _init_types()

    @staticmethod
    def get_type_index(typestr):
        type_property = Types.map.get(typestr)
        return type_property.index if type_property is not None else 11

    @staticmethod
    def is_builtin(typestr):
        return typestr in Types.map

    @staticmethod
    def is_collective(typestr):
        type_property = Types.map.get(typestr)
        return type_property.is_collective if type_property is not None else False

    @staticmethod
    def is_primitive(typestr):
        type_property = Types.map.get(typestr)
        return type_property.is_primitive if type_property is not None else False

    @staticmethod
    def parse(s):
        typespec, index = Types._parse_typespec(s, 0)
        return typespec

    @staticmethod
    def _parse_typespec(s, index):
        typestr = None
        details = []
        back_margin = 0
        start = index
        length = len(s)
        while (index < length):
            c = s[index]
            if c == '(' and index < (length - 1):
                typestr = s[start:index].strip()
                index += 1
                details, index = Types._parse_details(s, index)
                back_margin = 1
                break
            elif c == ',':
                index += 1
                back_margin = 1
                break
            elif c == ')':
                break
            index += 1

        if typestr is None:
            typestr = s[start:index - back_margin].strip()
        typespec = None if len(typestr) == 0 else TypeSpec(typestr, details)
        return typespec, index

    @staticmethod
    def _parse_details(s, index):
        details = []
        length = len(s)
        while (index < length):
            c = s[index]
            if c == ',':
                continue
            if c == ')':
                index += 1
                break
            else:
                detail, index = Types._parse_typespec(s, index)
                if detail is not None:
                    details.append(detail)
                    index -= 1
            index += 1
        if len(details) == 0:
            details = None
        return details, index



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import asyncio
from threading import Thread

from x2py.util.trace import Trace
from x2py.links.server_link import ServerLink
from x2py.links.asyncio.tcp_session import TcpSession

class TcpServer(ServerLink):
    def __init__(self, name):
        super(TcpServer, self).__init__(name)
        self.loop = asyncio.new_event_loop()
        self.server = None
        self.thread = Thread(target=self.loop.run_forever)

    def cleanup(self):
        self.loop.call_soon_threadsafe(self.loop.stop)

        self.thread.join()

        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())

        #self.loop.close()

        super(TcpServer, self).cleanup()

    def listen(self, host, port):
        factory = self.loop.create_server(self, host, port)
        self.server = self.loop.run_until_complete(factory)

        Trace.info("listening on {}:{}", host, port)

        self.thread.start()

    def _on_connect(self, result, context):
        super(TcpServer, self)._on_connect(result, context)
        if result:
            peername = context.transport.get_extra_info('peername')
            Trace.info("accepted from {}:{}", peername[0], peername[1])

    def _on_disconnect(self, handle, context):
        super(TcpServer, self)._on_disconnect(handle, context)
        peername = context.transport.get_extra_info('peername')
        Trace.info("disconnected from {}:{}", peername[0], peername[1])

    # protocol factory
    def __call__(self):
        return TcpSession(self)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from __future__ import print_function

import getopt
import sys

DEFAULT_SPEC = "py"

class Options(object):
    def __init__(self):
        self.forced = False
        self.out_dir = None
        self.recursive = False
        self.spec = DEFAULT_SPEC

    def parse(self, argv):
        longopts = ["spec=", "out-dir=", "recursive", "force", "help"]
        try:
            opts, args = getopt.getopt(argv, "s:o:rfh", longopts)
        except  getopt.GetoptError as err:
            print(err, file=sys.stderr)
            Options._usage()
            sys.exit(2)

        for opt, optarg in opts:
            if opt in ("-s", "--spec"):
                self.spec = optarg.lower()
                from main import Main
                if spec not in Main.formatters:
                    print("error: unknown target formatter specified - " + spec,
                        file=sys.stderr)
                    sys.exit(2)
            elif opt in ("-o", "--out-dir"):
                self.out_dir = optarg
            elif opt in ("-r", "--recursive"):
                self.recursive = True
            elif opt in ("-f", "--force"):
                self.forced = True
            elif opt in ("-h", "--help"):
                Options._usage()
                sys.exit(2)

        return args

    @staticmethod
    def _usage():
        print("usage: xpiler (options) [path...]")
        print(" options:")
        print("  -f (--force)       : force all to be re-xpiled")
        print("  -h (--help)        : print this message and quit")
        print("  -o (--out-dir) dir : specifies the output root directory")
        print("  -r (--recursive)   : process subdirectories recursively")
        print("  -s (--spec) spec   : specifies the target formatter")

        from main import Main
        for key, value in Main.formatters.items():
            suffix = " (default)" if key == DEFAULT_SPEC else ""
            line = "{:>20} : {}{}".format(key, value.desc(), suffix)
            print(line)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from abc import ABCMeta, abstractmethod

ABC = ABCMeta('ABC', (object,), {})

class BufferTransform(ABC):
    """Defines methods to be implemented by concrete buffer transforms."""

    @abstractmethod
    def cleanup(self):
        pass

    @abstractmethod
    def clone(self):
        pass

    @abstractmethod
    def handshake_block_length(self):
        pass

    @abstractmethod
    def init_handshake(self):
        pass
    @abstractmethod
    def handshake(self, challenge):
        pass
    @abstractmethod
    def fini_handshake(self, response):
        pass

    @abstractmethod
    def transform(self, buffer):
        """Transform the specified trailing byte(s) of the buffer."""
        pass
    @abstractmethod
    def inverse_transform(self, buffer):
        """Inverse transform the specified leading byte(s) of the buffer."""
        pass

class BufferTransformStack(BufferTransform):
    """Represents a single collective buffer transform that is actually a
        stacked set of child buffer transforms."""

    def __init__(self, transforms=None):
        self.transforms = []
        if transforms is None:
            return
        for transform in transforms:
            self.transforms.append(transform.clone())

    def add(self, transform):
        if transform not in self.transforms:
            self.transforms.append(transform)
        return self

    def remove(self, transform):
        if transform in self.transforms:
            self.transforms.remove(transform)
        return self

    def cleanup(self):
        for transform in self.transforms:
            transform.cleanup()
        del self.transforms[:]

    def clone(self):
        return BufferTransformStack(self.transforms)

    def handshake_block_length(self):
        result = 0
        for transform in self.transforms:
            result += transform.handshake_block_length()
        return result

    def init_handshake(self):
        result = bytearray()
        for transform in self.transforms:
            block = transform.init_handshake()
            if block:
                result += block
        return result

    def handshake(self, challenge):
        result = bytearray()
        offset = 0
        for transform in self.transforms:
            block_length = transform.handshake_block_length()
            if block_length:
                block = challenge[offset:offset + block_length]
                result += transform.handshake(block)
                offset += block_length
        return result

    def fini_handshake(self, response):
        offset = 0
        for transform in self.transforms:
            block_length = transform.handshake_block_length()
            if block_length:
                block = response[offset:offset + block_length]
                if not transform.fini_handshake(block):
                    return False
                offset += block_length
        return True

    def transform(self, buffer):
        for transform in self.transforms:
            buffer = transform.transform(buffer)
        return buffer

    def inverse_transform(self, buffer):
        for transform in reversed(self.transforms):
            buffer = transform.inverse_transform(buffer)
        return buffer



# auto-generated by x2py xpiler

from x2py.cell import MetaProperty, Cell
from x2py.event import Event

class BuiltinEventType(object):
    HEARTBEAT_EVENT = -1
    FLOW_START = -2
    FLOW_STOP = -3
    TIMEOUT_EVENT = -4
    PERIODIC_EVENT = -5

def _init_heartbeat_event_tag():
    props = []
    return Event.Tag(Event.tag, 'HeartbeatEvent', props,
        BuiltinEventType.HEARTBEAT_EVENT)

class HeartbeatEvent(Event):
    tag = _init_heartbeat_event_tag()

    def __init__(self, length=0):
        super(HeartbeatEvent, self).__init__(len(HeartbeatEvent.tag.props) + length)
        base = HeartbeatEvent.tag.offset
        pass

    def type_id(self):
        return HeartbeatEvent.tag.type_id

    def type_tag(self):
        return HeartbeatEvent.tag

def _init_flow_start_tag():
    props = []
    return Event.Tag(Event.tag, 'FlowStart', props,
        BuiltinEventType.FLOW_START)

class FlowStart(Event):
    tag = _init_flow_start_tag()

    def __init__(self, length=0):
        super(FlowStart, self).__init__(len(FlowStart.tag.props) + length)
        base = FlowStart.tag.offset
        pass

    def type_id(self):
        return FlowStart.tag.type_id

    def type_tag(self):
        return FlowStart.tag

def _init_flow_stop_tag():
    props = []
    return Event.Tag(Event.tag, 'FlowStop', props,
        BuiltinEventType.FLOW_STOP)

class FlowStop(Event):
    tag = _init_flow_stop_tag()

    def __init__(self, length=0):
        super(FlowStop, self).__init__(len(FlowStop.tag.props) + length)
        base = FlowStop.tag.offset
        pass

    def type_id(self):
        return FlowStop.tag.type_id

    def type_tag(self):
        return FlowStop.tag

def _init_timeout_event_tag():
    props = []
    props.append(MetaProperty('key', 15))
    props.append(MetaProperty('int_param', 5))
    return Event.Tag(Event.tag, 'TimeoutEvent', props,
        BuiltinEventType.TIMEOUT_EVENT)

class TimeoutEvent(Event):
    tag = _init_timeout_event_tag()

    def __init__(self, length=0):
        super(TimeoutEvent, self).__init__(len(TimeoutEvent.tag.props) + length)
        base = TimeoutEvent.tag.offset
        self.values[base + 0] = None
        self.values[base + 1] = 0

    @property
    def key(self):
        return self.values[TimeoutEvent.tag.offset + 0]
    @key.setter
    def key(self, value):
        self._set_property(TimeoutEvent.tag.offset + 0, value,
            TimeoutEvent.tag.props[0].type_index)

    @property
    def int_param(self):
        return self.values[TimeoutEvent.tag.offset + 1]
    @int_param.setter
    def int_param(self, value):
        self._set_property(TimeoutEvent.tag.offset + 1, value,
            TimeoutEvent.tag.props[1].type_index)

    def type_id(self):
        return TimeoutEvent.tag.type_id

    def type_tag(self):
        return TimeoutEvent.tag

def _init_periodic_event_tag():
    props = []
    props.append(MetaProperty('key', 15))
    props.append(MetaProperty('int_param', 5))
    return Event.Tag(Event.tag, 'PeriodicEvent', props,
        BuiltinEventType.PERIODIC_EVENT)

class PeriodicEvent(Event):
    tag = _init_periodic_event_tag()

    def __init__(self, length=0):
        super(PeriodicEvent, self).__init__(len(PeriodicEvent.tag.props) + length)
        base = PeriodicEvent.tag.offset
        self.values[base + 0] = None
        self.values[base + 1] = 0

    @property
    def key(self):
        return self.values[PeriodicEvent.tag.offset + 0]
    @key.setter
    def key(self, value):
        self._set_property(PeriodicEvent.tag.offset + 0, value,
            PeriodicEvent.tag.props[0].type_index)

    @property
    def int_param(self):
        return self.values[PeriodicEvent.tag.offset + 1]
    @int_param.setter
    def int_param(self, value):
        self._set_property(PeriodicEvent.tag.offset + 1, value,
            PeriodicEvent.tag.props[1].type_index)

    def type_id(self):
        return PeriodicEvent.tag.type_id

    def type_tag(self):
        return PeriodicEvent.tag



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.deserializer import Deserializer
from x2py.event_factory import EventFactory
from x2py.hub import Hub
from x2py.serializer import Serializer
from x2py.util.trace import Trace

from x2py.links.link_events import *

class LinkSession(object):
    """Abstract base class for concrete link sessions."""

    def __init__(self, link):  # link: SessionBasedLink
        self.link = link
        self.handle = 0
        self.polarity = False
        self.rx_buffer = bytearray()
        self.channel_strategy = None
        self.heartbeat_strategy = None

    @property
    def has_channel_strategy(self):
        return (self.channel_strategy is not None)

    @property
    def has_heartbeat_strategy(self):
        return (self.heartbeat_strategy is not None)

    def cleanup(self):
        if self.has_channel_strategy:
            self.channel_strategy.cleanup()

    def close(self):
        self.cleanup()

    def on_receive(self, data):
        if self.has_heartbeat_strategy:
            self.heartbeat_strategy.on_receive()

        self.rx_buffer += data

        deserializer = Deserializer()
        while True:
            deserializer.buffer = self.rx_buffer
            deserializer.pos = 0

            num_bytes, length, transformed = self.parse_header(deserializer)
            if num_bytes == 0:
                return

            if len(self.rx_buffer) < (length + num_bytes):
                return
            buffer = self.rx_buffer[num_bytes:num_bytes + length]
            self.rx_buffer = self.rx_buffer[num_bytes + length:]

            if self.has_channel_strategy and transformed:
                try:
                    buffer = self.channel_strategy.after_receive(buffer)
                except Exception as ex:
                    Trace.error("{} inverse transform error {}", self.link.name, ex)
                    continue

            deserializer.buffer = buffer
            deserializer.pos = 0

            type_id = deserializer.read_int32(None)

            event = self.link.create_event(type_id)
            if event is None:
                Trace.Error("unknown event type id {}", type_id)
                continue
            event.deserialize(deserializer)

            Trace.debug("{} received {}", self.link.name, event)

            processed = False

            if self.has_channel_strategy:
                processed = self.channel_strategy.process(event)
            elif not processed and self.has_heartbeat_strategy:
                processed = self.heartbeat_strategy.process(event)
            elif not processed:
                processed = self._process(event)

            if not processed:
                event._handle = self.handle
                Hub.post(event)

    def _process(self, event):
        return False

    def send(self, event):
        serializer = Serializer()
        serializer.write_int32(None, event.type_id())
        event.serialize(serializer)
        buffer = serializer.buffer

        transformed = False
        if self.has_channel_strategy and event._transform:
            transformed, buffer = self.channel_strategy.before_send(buffer)

        header_buffer = self.build_header(buffer, transformed)

        data = bytes(header_buffer + buffer)

        if self.has_heartbeat_strategy:
            self.heartbeat_strategy.on_send(event)

        self._send(data)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import asyncore
import errno
import socket
from threading import Thread

from x2py.util.trace import Trace
from x2py.links.client_link import ClientLink
from x2py.links.asyncore.tcp_session import TcpSession

class TcpClient(ClientLink):
    """TCP/IP client link based on the asyncore module."""

    class Dispatcher(asyncore.dispatcher):
        def __init__(self, owner):
            asyncore.dispatcher.__init__(self, map=owner.map)
            self.owner = owner
        def handle_connect(self):
            self.owner.handle_connect()
        def handle_close(self):
            self.owner.handle_close()
        def handle_error(self):
            self.owner.handle_error()

    def __init__(self, name):
        super(TcpClient, self).__init__(name)
        self.map = {}
        self.dispatcher = TcpClient.Dispatcher(self)
        self.thread = Thread(target=self._loop)
        self.session = None
        self.remote_host = ''
        self.remote_port = 0
        self.connecting = False

    def cleanup(self):
        asyncore.close_all(map=self.map)

        self.thread.join()

        super(TcpClient, self).cleanup()

    def connect(self, host, port):
        self.connecting = True
        self.remote_host = host
        self.remote_port = port
        Trace.info("connecting to {}:{}", host, port)
        self.dispatcher.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.dispatcher.connect((host, port))

        self.thread.start()

    def handle_connect(self):
        self.sesison = TcpSession(self, self.dispatcher.socket)
        self.sesison.connection_made()

    def handle_close(self):
        self.handle_error()
        self.dispatcher.close()

    def handle_error(self):
        err = self.dispatcher.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if self.connecting:
            Trace.error("connect error {}", errno.errorcode[err])
            self.connecting = False
            return
        Trace.error("error {}", errno.errorcode[err])

    def _loop(self):
        asyncore.loop(map=self.map)

    def _on_connect(self, result, context):
        super(TcpClient, self)._on_connect(result, context)
        if result:
            peername = context.socket.getpeername()
            Trace.info("connected to {}:{}", peername[0], peername[1])
            context.peername = peername
        else:
            Trace.error("error connecting to {}:{}", self.remote_host, self.remote_port)

    def _on_disconnect(self, handle, context):
        super(TcpClient, self)._on_disconnect(handle, context)
        peername = context.peername
        Trace.info("disconnected from {}:{}", peername[0], peername[1])



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.coroutine import Coroutine

class WaitForCompletion(object):
    def __init__(self, coroutine, routine):
        c = Coroutine(coroutine)
        c.start(routine(c))



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import asyncore
import socket

from x2py.deserializer import Deserializer
from x2py.serializer import Serializer
from x2py.util.trace import Trace
from x2py.links.link_session import LinkSession

class TcpSession(LinkSession):
    """TCP/IP link session based on the asyncore module."""

    class Dispatcher(asyncore.dispatcher_with_send):
        def __init__(self, owner, sock, map):
            asyncore.dispatcher_with_send.__init__(self, sock, map)
            self.owner = owner
        def handle_read(self):
            self.owner.handle_read()

    def __init__(self, link, sock):
        super(TcpSession, self).__init__(link)
        self.dispatcher = TcpSession.Dispatcher(self, sock, link.map)
        self.socket = sock

        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)

    def cleanup(self):
        super(TcpSession, self).cleanup()
        self.dispatcher.close()

    def connected(self):
        if self.socket is not None:
            return (self.socket.fileno() >= 0)
        return False

    def handle_read(self):
        data = self.dispatcher.recv(4096)
        if not data:
            self.connection_lost()
            return
        Trace.trace("{} received {}", self.link.name, repr(data))
        self.on_receive(data)

    def connection_made(self):
        self.link.init_session(self)

    def connection_lost(self):
        self.link.on_disconnect(self.handle, self)

    def build_header(self, buffer, transformed):
        length = len(buffer)
        header = 1 if transformed else 0
        header = header | (length << 1)

        result = bytearray()
        Serializer.write_variable(result, header)
        return result

    def parse_header(self, deserializer):
        pos = deserializer.pos
        try:
            header, num_bytes = deserializer.read_variable32()
        except:
            return 0, 0, False
        finally:
            deserializer.pos = pos

        length = header >> 1
        transformed = ((header & 1) != 0)
        return num_bytes, length, transformed

    def _send(self, data):
        Trace.trace("{} sending {}", self.link.name, repr(data))
        self.dispatcher.send(data)



# auto-generated by x2py xpiler

from x2py.cell import MetaProperty, Cell
from x2py.event import Event

class LinkEventType(object):
    LINK_SESSION_CONNECTED = -10
    LINK_SESSION_DISCONNECTED = -11
    HANDSHAKE_REQ = -12
    HANDSHAKE_RESP = -13
    HANDSHAKE_ACK = -14

def _init_link_session_connected_tag():
    props = []
    props.append(MetaProperty('link_name', 9))
    props.append(MetaProperty('result', 1))
    props.append(MetaProperty('context', 15))
    return Event.Tag(Event.tag, 'LinkSessionConnected', props,
        LinkEventType.LINK_SESSION_CONNECTED)

class LinkSessionConnected(Event):
    tag = _init_link_session_connected_tag()

    def __init__(self, length=0):
        super(LinkSessionConnected, self).__init__(len(LinkSessionConnected.tag.props) + length)
        base = LinkSessionConnected.tag.offset
        self.values[base + 0] = ""
        self.values[base + 1] = False
        self.values[base + 2] = None

    @property
    def link_name(self):
        return self.values[LinkSessionConnected.tag.offset + 0]
    @link_name.setter
    def link_name(self, value):
        self._set_property(LinkSessionConnected.tag.offset + 0, value,
            LinkSessionConnected.tag.props[0].type_index)

    @property
    def result(self):
        return self.values[LinkSessionConnected.tag.offset + 1]
    @result.setter
    def result(self, value):
        self._set_property(LinkSessionConnected.tag.offset + 1, value,
            LinkSessionConnected.tag.props[1].type_index)

    @property
    def context(self):
        return self.values[LinkSessionConnected.tag.offset + 2]
    @context.setter
    def context(self, value):
        self._set_property(LinkSessionConnected.tag.offset + 2, value,
            LinkSessionConnected.tag.props[2].type_index)

    def type_id(self):
        return LinkSessionConnected.tag.type_id

    def type_tag(self):
        return LinkSessionConnected.tag

def _init_link_session_disconnected_tag():
    props = []
    props.append(MetaProperty('link_name', 9))
    props.append(MetaProperty('handle', 5))
    props.append(MetaProperty('context', 15))
    return Event.Tag(Event.tag, 'LinkSessionDisconnected', props,
        LinkEventType.LINK_SESSION_DISCONNECTED)

class LinkSessionDisconnected(Event):
    tag = _init_link_session_disconnected_tag()

    def __init__(self, length=0):
        super(LinkSessionDisconnected, self).__init__(len(LinkSessionDisconnected.tag.props) + length)
        base = LinkSessionDisconnected.tag.offset
        self.values[base + 0] = ""
        self.values[base + 1] = 0
        self.values[base + 2] = None

    @property
    def link_name(self):
        return self.values[LinkSessionDisconnected.tag.offset + 0]
    @link_name.setter
    def link_name(self, value):
        self._set_property(LinkSessionDisconnected.tag.offset + 0, value,
            LinkSessionDisconnected.tag.props[0].type_index)

    @property
    def handle(self):
        return self.values[LinkSessionDisconnected.tag.offset + 1]
    @handle.setter
    def handle(self, value):
        self._set_property(LinkSessionDisconnected.tag.offset + 1, value,
            LinkSessionDisconnected.tag.props[1].type_index)

    @property
    def context(self):
        return self.values[LinkSessionDisconnected.tag.offset + 2]
    @context.setter
    def context(self, value):
        self._set_property(LinkSessionDisconnected.tag.offset + 2, value,
            LinkSessionDisconnected.tag.props[2].type_index)

    def type_id(self):
        return LinkSessionDisconnected.tag.type_id

    def type_tag(self):
        return LinkSessionDisconnected.tag

def _init_handshake_req_tag():
    props = []
    props.append(MetaProperty('data', 11))
    return Event.Tag(Event.tag, 'HandshakeReq', props,
        LinkEventType.HANDSHAKE_REQ)

class HandshakeReq(Event):
    tag = _init_handshake_req_tag()

    def __init__(self, length=0):
        super(HandshakeReq, self).__init__(len(HandshakeReq.tag.props) + length)
        base = HandshakeReq.tag.offset
        self.values[base + 0] = None

    @property
    def data(self):
        return self.values[HandshakeReq.tag.offset + 0]
    @data.setter
    def data(self, value):
        self._set_property(HandshakeReq.tag.offset + 0, value,
            HandshakeReq.tag.props[0].type_index)

    def type_id(self):
        return HandshakeReq.tag.type_id

    def type_tag(self):
        return HandshakeReq.tag

def _init_handshake_resp_tag():
    props = []
    props.append(MetaProperty('data', 11))
    return Event.Tag(Event.tag, 'HandshakeResp', props,
        LinkEventType.HANDSHAKE_RESP)

class HandshakeResp(Event):
    tag = _init_handshake_resp_tag()

    def __init__(self, length=0):
        super(HandshakeResp, self).__init__(len(HandshakeResp.tag.props) + length)
        base = HandshakeResp.tag.offset
        self.values[base + 0] = None

    @property
    def data(self):
        return self.values[HandshakeResp.tag.offset + 0]
    @data.setter
    def data(self, value):
        self._set_property(HandshakeResp.tag.offset + 0, value,
            HandshakeResp.tag.props[0].type_index)

    def type_id(self):
        return HandshakeResp.tag.type_id

    def type_tag(self):
        return HandshakeResp.tag

def _init_handshake_ack_tag():
    props = []
    props.append(MetaProperty('result', 1))
    return Event.Tag(Event.tag, 'HandshakeAck', props,
        LinkEventType.HANDSHAKE_ACK)

class HandshakeAck(Event):
    tag = _init_handshake_ack_tag()

    def __init__(self, length=0):
        super(HandshakeAck, self).__init__(len(HandshakeAck.tag.props) + length)
        base = HandshakeAck.tag.offset
        self.values[base + 0] = False

    @property
    def result(self):
        return self.values[HandshakeAck.tag.offset + 0]
    @result.setter
    def result(self, value):
        self._set_property(HandshakeAck.tag.offset + 0, value,
            HandshakeAck.tag.props[0].type_index)

    def type_id(self):
        return HandshakeAck.tag.type_id

    def type_tag(self):
        return HandshakeAck.tag



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from datetime import timedelta

from x2py.builtin_events import HeartbeatEvent
from x2py.config import Config
from x2py.event import Event
from x2py.flow import Flow
from x2py.flows.time_flow import TimeFlow
from x2py.util.rwlock import ReadLock, WriteLock, ReadWriteLock
from x2py.util.trace import Trace

class Hub(object):
    """Represents the singleton event distribution bus."""

    class _Hub(object):
        def __init__(self):
            self.cases = []  # list of hub cases
            self.flows = []  # list of all the flows attached to this hub
            self.rwlock = ReadWriteLock()

        def add(self, case):
            """Adds the specified hub case to the hub."""
            if case is None or not isinstance(case, Hub.Case):
                raise TypeError()
            with self.rwlock.wlock():
                if case not in self.cases:
                    self.cases.append(case)
                    Trace.debug("hub: added case {}", type(case).__name__)
            return self

        def insert(self, index, case):
            """Inserts the specified hub case to the hub, at the specified index."""
            if case is None or not isinstance(case, Hub.Case):
                raise TypeError()
            with self.rwlock.wlock():
                if case not in self.cases:
                    self.cases.insert(index, case)
                    Trace.debug("hub: inserted case {} at index {}",
                        type(case).__name__, index)
            return self

        def remove(self, case):
            """Removes the specified hub case from the hub."""
            if case is None or not isinstance(case, Hub.Case):
                raise TypeError()
            with self.rwlock.wlock():
                if case in self.cases:
                    self.cases.remove(case)
                    Trace.debug("hub: removed case {}", type(case).__name__)
            return self

        def attach(self, flow):
            """Attaches the specified flow to the hub."""
            if flow is None or not isinstance(flow, Flow):
                raise TypeError()
            with self.rwlock.wlock():
                if flow not in self.flows:
                    self.flows.append(flow)
                    Trace.debug("hub: attached flow {}", flow.name)
            return self

        def detach(self, flow):
            """Detaches the specified flow from the hub."""
            if flow is None or not isinstance(flow, Flow):
                raise TypeError()
            with self.rwlock.wlock():
                if flow in self.flows:
                    self.flows.remove(flow)
                    Trace.debug("hub: detached flow {}", flow.name)
            return self

        def detach_all(self):
            """Detaches all the attached flows."""
            snapshot = self.flows[::-1]
            for flow in snapshot:
                self.detach(flow)

        def feed(self, event):
            if event is None or not isinstance(event, Event):
                raise TypeError()
            with self.rwlock.rlock():
                for flow in self.flows:
                    flow.feed(event)

        def setup(self):
            with self.rwlock.rlock():
                snapshot = self.cases[:]
            for case in snapshot:
                Trace.trace("hub: setting up case {}", type(case).__name__)
                case.setup()

        def teardown(self):
            with self.rwlock.rlock():
                snapshot = self.cases[::-1]
            for case in snapshot:
                try:
                    Trace.trace("hub: tearing down case {}", type(case).__name__)
                    case.teardown()
                except:
                    pass

        def start_flows(self):
            with self.rwlock.rlock():
                snapshot = self.flows[:]
            for flow in snapshot:
                Trace.trace("hub: starting flow {}", flow.name)
                flow.start()
                Trace.debug("hub: started flow {}", flow.name)

        def stop_flows(self):
            with self.rwlock.rlock():
                snapshot = self.flows[::-1]
            for flow in snapshot:
                try:
                    Trace.trace("hub: stopping flow {}", flow.name)
                    flow.stop()
                    Trace.debug("hub: stopped flow {}", flow.name)
                except:
                    pass

    instance = _Hub()
    heartbeat_event = HeartbeatEvent().setattrs( _transform = False )

    def __init__(self):
        raise AssertionError()

    @staticmethod
    def post(event):
        """Posts up the specified event to the hub."""
        Hub.instance.feed(event)

    @staticmethod
    def startup():
        """Starts all the flows attached to the hub."""
        Trace.debug("starting up")
        Hub.instance.setup()
        Hub.instance.start_flows()
        TimeFlow.get().reserve_repetition(Hub.heartbeat_event, \
            timedelta(seconds=Config.heartbeat_interval))
        Trace.info("started")

    @staticmethod
    def shutdown():
        """Stops all the flows attached to the hub."""
        Trace.debug("shutting down")
        Hub.instance.stop_flows()
        Hub.instance.teardown()
        Trace.info("stopped")

    class Case(object):
        """Represents a hub-scope case that is initialized and terminated
        along with startup/shutdown of the hub."""

        def setup(self):
            """Overridden by subclasses for initialization."""
            pass

        def teardown(self):
            """Overridden by subclasses for clean-up."""
            pass

    class Flows(object):
        """Represents the set of attached flows for convenient cleanup."""

        def startup(self):
            """Starts all the attached flows."""
            Hub.startup()

        def shutdown(self):
            """Stops all the attached flows."""
            Hub.shutdown()

        def __enter__(self):
            self.startup()
            return self

        def __exit__(self, type, value, traceback):
            self.shutdown()



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from __future__ import print_function

import sys
import xml.etree.ElementTree as etree

from xpiler.definition import *
from xpiler.handler import Handler
from xpiler.types_ import *

class XmlHandler(Handler):
    def handle(self, path):
        unit = None
        try:
            tree = etree.parse(path)
        except etree.ParseError as err:
            print(err, file=sys.stderr)
            return False, unit

        root = tree.getroot()
        if root.tag != 'x2':
            # Not a valid x2 document.
            return True, unit

        return self._normalize(root)

    def _normalize(self, root):
        unit = Unit()
        if 'namespace' in root.attrib:
            unit.namespace = root.attrib['namespace']
        if 'builtin' in root.attrib and root.attrib['builtin'].endswith('rue'):
            unit.is_builtin = True

        for node in root:
            if node.tag == 'references':
                if self._parse_references(unit, node) == False:
                    return False, unit
            elif node.tag == 'definitions':
                if self._parse_definitions(unit, node) == False:
                    return False, unit

        return True, unit

    def _parse_references(self, unit, node):
        for child in node:
            # Counts file references only
            if child.tag == 'file':
                ref = Reference(child.attrib['target'].strip())
                unit.references.append(ref)

    def _parse_definitions(self, unit, node):
        for child in node:
            if child.tag == 'consts':
                if self._parse_consts(unit, child) == False:
                    return False
            elif child.tag in ('cell', 'event'):
                if self._parse_cell(unit, child) == False:
                    return False

    def _parse_consts(self, unit, node):
        if 'name' not in node.attrib:
            return False
        name = node.attrib['name']
        if len(name) == 0:
            return False

        typestr = node.attrib['type'] if 'type' in node.attrib else 'int32'

        consts = Consts(name, typestr)

        for child in node:
            if child.tag == 'const':
                if self._parse_const(consts, child) == False:
                    return False

        unit.definitions.append(consts)
        return True

    def _parse_const(self, consts, node):
        if 'name' not in node.attrib:
            return False
        name = node.attrib['name']
        if len(name) == 0:
            return False

        const = Constant(name, node.text)
        consts.constants.append(const)
        return True

    def _parse_cell(self, unit, node):
        if 'name' not in node.attrib:
            return False
        name = node.attrib['name']
        if len(name) == 0:
            return False

        is_event = (node.tag == 'event')
        if is_event:
            if 'id' not in node.attrib:
                return False
            id = node.attrib['id']
            if len(id) == 0:
                return False

        cell = Event() if is_event else Cell()
        cell.name = name
        if is_event:
            cell.id = id
        cell.base = node.attrib['base'] if 'base' in node.attrib else ''

        if 'local' in node.attrib and node.attrib['local'].endswith('rue'):
            cell.local = True

        for child in node:
            if child.tag == 'property':
                if self._parse_property(cell, child) == False:
                    return False

        unit.definitions.append(cell)
        return True

    def _parse_property(self, cell, node):
        if ('name' not in node.attrib) or ('type' not in node.attrib):
            return False
        name = node.attrib['name']
        typestr = node.attrib['type']
        if len(name) == 0 or len(typestr) == 0:
            return False

        typespec = Types.parse(typestr)
        if typespec is None:
            return False

        default_value = node.text.strip() if node.text is not None else ''

        prop = Property(name, default_value)
        prop.typespec = typespec
        cell.properties.append(prop)
        return True



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import time
from threading import Lock, Thread

from x2py.event import Event
from x2py.event_queue import EventQueue
from x2py.flow import Flow
from x2py.util.atomic import AtomicInt
from x2py.util.trace import Trace

class Time(object):
    """Utility class to handle time information within a frame-based flow."""

    def __init__(self):
        pass

    def init(self):
        self.start_clock = Time.clock()
        self.last_clock = self.start_clock
        self.current_clock = self.last_clock
        self.delta_clock = 0.0

    def before_update(self):
        self.current_clock = Time.clock()
        self.delta_clock = self.current_clock - self.last_clock

    def after_update(self):
        self.last_clock = self.current_clock

    @staticmethod
    def clock():
        return time.clock()

class FrameBasedFlow(Flow):
    """Abstract base class for frame-based (looping) execution flows."""

    def __init__(self, name=None, with_queue=False):
        super(FrameBasedFlow, self).__init__(name)
        self.queue = None
        if with_queue:
            self.queue = EventQueue()
        self._lock = Lock()
        self.should_stop = AtomicInt()
        self.thread = None

        # Default resolution is 15.625ms (64 frame/sec)
        self.resolution = 0.015625

        self.time = Time()

    def feed(self, event):
        if self.queue is None:
            return
        if event is None or not isinstance(event, Event):
            raise TypeError()
        self.queue.enqueue(event)

    def start(self):
        with self._lock:
            if self.thread is not None:
                return

            self._setup()
            self.cases.setup_with(self)

            self.thread = Thread(target=self)
            if self.name is not None:
                self.thread.setName(self.name)
            self.thread.start()

            if self.queue is not None:
                self.queue.enqueue(FlowStart())

    def stop(self):
        with self._lock:
            if self.thread is None:
                return

            self.should_stop.value = 1
            if self.queue is not None:
                self.queue.close(FlowStop())

            self.thread.join()
            self.thread = None

            self.cases.teardown_with(self)
            self._teardown()

    def __call__(self):
        Flow.thread_local.current = self
        if self.queue is not None:
            Flow.thread_local.event_proxy = EventProxy()
            Flow.thread_local.handler_chain = []

        self._begin()

        while self.should_stop.value == 0:
            self._update()

            if self.queue is not None:
                while (Time.clock() - self.time.current_clock) < self.resolution:
                    event = self.queue.try_dequeue()
                    if event is not None:
                        self.dispatch(event)
                        if event.type_id() == BuiltinEventType.FLOW_STOP:
                            self.should_stop.value = 1
                            break
                    else:
                        if self.should_stop.value != 0:
                            break
                        else:
                            time.sleep(0.001)
                            continue
            else:
                clock_delta = Time.clock() - self.time.current_clock
                if clock_delta < self.resolution:
                    delay = self.resolution - clock_delta
                else:
                    delay = 0.0
                time.sleep(delay)

        self.end()

        if self.queue is not None:
            Flow.thread_local.handler_chain = None
            Flow.thread_local.event_proxy = None
        Flow.thread_local.current = None

    def _begin(self):
        self.time.init()
        self.begin()

    def _update(self):
        self.time.before_update()
        self.update()
        self.time.after_update()

    def begin(self):
        pass

    def end(self):
        pass

    def update(self):
        raise NotImplementedError()



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.links.session_based_link import SessionBasedLink

class ClientLink(SessionBasedLink):
    """Common base class for single-session client links."""

    def __init__(self, name):
        super(ClientLink, self).__init__(name)
        self.session = None  # current link session

    def cleanup(self):
        if self.session is None:
            return
        self.session.close()
        self.sesison = None
        super(ClientLink, self).cleanup()

    def send(self, event):
        if self.session is not None:
            self.session.send(event)

    def _on_connect(self, result, context):
        super(ClientLink, self)._on_connect(result, context)
        if result:
            context.polarity = True
            self.session = context

    def _on_disconnect(self, handle, context):
        super(ClientLink, self)._on_disconnect(handle, context)
        self.session = None



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.builtin_events import *
from x2py.event import EventProxy
from x2py.flow import Flow
from x2py.flows.event_based_flow import EventBasedFlow
from x2py.util.trace import Trace

class ThreadlessFlow(EventBasedFlow):
    def __init__(self, name=None):
        super(ThreadlessFlow, self).__init__(name)
        self.running = False

    def start(self):
        with self._lock:
            if self.running:
                return

            self._setup()
            self.cases.setup_with(self)

            Flow.thread_local.current = self
            Flow.thread_local.event_proxy = EventProxy()
            Flow.thread_local.handler_chain = []

            self.running = True
            self.queue.enqueue(FlowStart())

    def stop(self):
        with self._lock:
            if not self.running:
                return

            self.queue.close(FlowStop())
            self.running = False

            Flow.thread_local.handler_chain = None
            Flow.thread_local.event_proxy = None
            Flow.thread_local.current = None

            self.cases.teardown_with(self)
            self._teardown()

    def dispatch(self):
        event = self.queue.dequeue()
        if event is None:
            return
        self.dispatch(event)

    def try_dispatch(self):
        event = self.queue.try_dequeue()
        if event is not None:
            self.dispatch(event)
        return event

    def try_dispatch_all(self):
        n = 0
        while True:
            event = self.queue.try_dequeue()
            if event is None:
                break
            self.dispatch(event)
            n += 1
        return n



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.links.session_based_link import SessionBasedLink

class ServerLink(SessionBasedLink):
    def __init__(self, name):
        super(ServerLink, self).__init__(name)
        self.sessions = {}

    def cleanup(self):
        super(ServerLink, self).cleanup()

    def send(self, event):
        handle = event._handle
        if handle == 0:
            return
        with self.rwlock.rlock():
            session = self.sessions.get(handle)
        if session is not None:
            session.send(event)

    def _on_connect(self, result, context):
        super(ServerLink, self)._on_connect(result, context)
        if result:
            session = context
            with self.rwlock.wlock():
                self.sessions[session.handle] = session

    def _on_disconnect(self, handle, context):
        super(ServerLink, self)._on_disconnect(handle, context)
        session = context
        with self.rwlock.wlock():
            self.sessions.pop(session.handle)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import datetime
import struct

class Serializer(object):
    """Binary wire format serializer."""

    @staticmethod
    def length_bool(metaprop, value):
        return 1

    def write_bool(self, metaprop, value):
        if value:
            self.buffer.append(1)
        else:
            self.buffer.append(0)

    @staticmethod
    def length_byte(metaprop, value):
        return 1

    def write_byte(self, metaprop, value):
        self.buffer.append(value)

    @staticmethod
    def length_int8(metaprop, value):
        return 1

    def write_int8(self, metaprop, value):
        self.buffer += struct.pack('b', value)

    @staticmethod
    def length_int16(metaprop, value):
        return 2

    def write_int16(self, metaprop, value):
        self.buffer += struct.pack('!h', value)

    @staticmethod
    def length_int32(metaprop, value):
        return Serializer.len_variable32(value)

    def write_int32(self, metaprop, value):
        value = (value << 1) ^ (value >> 31)
        Serializer.write_variable(self.buffer, value)

    @staticmethod
    def length_int64(metaprop, value):
        return Serializer.len_variable64(value)

    def write_int64(self, metaprop, value):
        value = (value << 1) ^ (value >> 63)
        Serializer.write_variable(self.buffer, value)

    @staticmethod
    def length_float32(metaprop, value):
        return 4

    def write_float32(self, metaprop, value):
        self.buffer += struct.pack('!f', value)

    @staticmethod
    def length_float64(metaprop, value):
        return 8

    def write_float64(self, metaprop, value):
        self.buffer += struct.pack('!d', value)

    @staticmethod
    def length_string(metaprop, value):
        length = Serializer.length_utf8(value)
        return Serializer.get_length_nonnegative(length) + length

    @staticmethod
    def length_utf8(value):
        length = 0
        if value is not None:
            for char in value:
                c = ord(char)
                if (c & 0xff80) == 0:
                    length += 1
                elif (c & 0xf800) != 0:
                    length += 3
                else:
                    length += 2
        return length

    def write_string(self, metaprop, value):
        # utf-8 encoding
        length = Serializer.length_utf8(value)
        Serializer.write_variable(self.buffer, length)  # write_nonnegative
        if length == 0:
            return
        for char in value:
            c = ord(char)
            if (c & 0xff80) == 0:
                self.buffer.append(c & 0x0ff)
            elif (c & 0xf800) != 0:
                self.buffer.append(0x0e0 | ((c >> 12) & 0x0f))
                self.buffer.append(0x080 | ((c >> 6) & 0x3f))
                self.buffer.append(0x080 | ((c >> 0) & 0x3f))
            else:
                self.buffer.append(0x0c0 | ((c >> 6) & 0x1f))
                self.buffer.append(0x080 | ((c >> 0) & 0x3f))

    @staticmethod
    def length_datetime(metaprop, value):
        return 8

    def write_datetime(self, metaprop, value):
        unix_epoch = datetime.datetime(1970, 1, 1)
        millisecs = int((value - unix_epoch).total_seconds() * 1000)
        self.buffer += struct.pack('!q', millisecs)

    @staticmethod
    def length_bytes(metaprop, value):
        length = 0 if value is None else len(value)
        return Serializer.get_length_nonnegative(length) + length

    def write_bytes(self, metaprop, value):
        is_none = (value is None)
        length = 0 if is_none else len(value)
        Serializer.write_variable(self.buffer, length)  # write_nonnegative
        if is_none:
            return
        self.buffer += value

    @staticmethod
    def length_cell(metaprop, value):
        from x2py.event import Event
        if value is None:
            return 1
        length = 0
        partial = False
        if isinstance(value, Event):
            length = Serializer.length_int32(None, value.type_id())
        elif metaprop.runtime_type != type(value):
            partial = True
        if partial:
            length += value.get_length(metaprop.runtime_type)
        else:
            length += value.get_length()
        return Serializer.get_length_nonnegative(length) + length

    def write_cell(self, metaprop, value):
        from x2py.event import Event
        if value is None:
            Serializer.write_variable(self.buffer, 0)  # write_nonnegative
            return
        length = 0
        partial = False
        is_event = isinstance(value, Event)
        if is_event:
            type_id = value.type_id()
            length = Serializer.length_int32(None, type_id)
        elif metaprop.runtime_type != type(value):
            partial = True
        if partial:
            length += value.get_length(metaprop.runtime_type)
        else:
            length += value.get_length()
        Serializer.write_variable(self.buffer, length)  # write_nonnegative
        if is_event:
            self.write_int32(None, type_id)
        if partial:
            value.serialize(self, metaprop.runtime_type)
        else:
            value.serialize(self)

    @staticmethod
    def length_list(metaprop, value):
        is_none = (value is None)
        length = 0 if is_none else len(value)
        result = Serializer.get_length_nonnegative(length)
        for v in value:
            result += Serializer.get_length(metaprop.details[0], v)
        return result

    def write_list(self, metaprop, value):
        is_none = (value is None)
        length = 0 if is_none else len(value)
        Serializer.write_variable(self.buffer, length)  # write_nonnegative
        if is_none:
            return
        for v in value:
            self.write(metaprop.details[0], v)

    @staticmethod
    def length_map(metaprop, value):
        is_none = (value is None)
        length = 0 if is_none else len(value)
        result = Serializer.get_length_nonnegative(length)
        for k, v in value.items():
            result += Serializer.get_length(metaprop.details[0], k)
            result += Serializer.get_length(metaprop.details[1], v)
        return result

    def write_map(self, metaprop, value):
        is_none = (value is None)
        length = 0 if is_none else len(value)
        Serializer.write_variable(self.buffer, length)  # write_nonnegative
        if is_none:
            return
        for k, v in value.items():
            self.write(metaprop.details[0], k)
            self.write(metaprop.details[1], v)

    # Length estimation function table
    lengths = [
        None,
        length_bool,
        length_byte,
        length_int8,
        length_int16,
        length_int32,
        length_int64,
        length_float32,
        length_float64,
        length_string,
        length_datetime,
        length_bytes,
        length_cell,
        length_list,
        length_map,
        None  # none for object type
    ]
    # Writer function table
    writers = [
        None,
        write_bool,
        write_byte,
        write_int8,
        write_int16,
        write_int32,
        write_int64,
        write_float32,
        write_float64,
        write_string,
        write_datetime,
        write_bytes,
        write_cell,
        write_list,
        write_map,
        None  # none for object type
    ]

    def __init__(self, buffer=None):
        self.buffer = buffer
        if self.buffer is None:
            self.buffer = bytearray()

    @staticmethod
    def get_length(metaprop, value):
        length = Serializer.lengths[metaprop.type_index]
        return length.__func__(metaprop, value)

    def write(self, metaprop, value):
        writer = Serializer.writers[metaprop.type_index]
        writer(self, metaprop, value)

    @staticmethod
    def get_length_nonnegative(value):
        if value < 0:
            raise ValueError()
        return Serializer.len_variable32(value)

    def write_nonnegative(self, value):
        if value < 0:
            raise ValueError()
        Serializer.write_variable(self.buffer, value)

    @staticmethod
    def len_variable32(value):
        if (value & 0xffffff80) == 0:
            return 1
        if (value & 0xffffc000) == 0:
            return 2
        if (value & 0xffe00000) == 0:
            return 3
        if (value & 0xf0000000) == 0:
            return 4
        return 5

    @staticmethod
    def len_variable64(value):
        if (value & 0xffffffffffffff80) == 0:
            return 1
        if (value & 0xffffffffffffc000) == 0:
            return 2
        if (value & 0xffffffffffe00000) == 0:
            return 3
        if (value & 0xfffffffff0000000) == 0:
            return 4
        if (value & 0xfffffff800000000) == 0:
            return 5
        if (value & 0xfffffc0000000000) == 0:
            return 6
        if (value & 0xfffe000000000000) == 0:
            return 7
        if (value & 0xff00000000000000) == 0:
            return 8
        if (value & 0x8000000000000000) == 0:
            return 9
        return 10

    @staticmethod
    def write_variable(buffer, value):
        while True:
            b = value & 0x7f
            value = value >> 7
            if value != 0:
                b = b | 0x80
            buffer.append(b)
            if value == 0:
                break



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

class Coroutine(object):
    def __init__(self, parent=None):
        self.result = None
        self.routine = None
        self.running = False
        self.started = False
        self.parent = parent

    @staticmethod
    def start(routine):
        coroutine = Coroutine()
        coroutine.start(routine(coroutine))

    @staticmethod
    def start1(routine, arg1):
        coroutine = Coroutine()
        coroutine.start(routine(coroutine, arg1))

    def start(self, routine):
        self.routine = iter(routine)
        self.running = bool(routine)
        if not self.next():
            if self.parent:
                # Indirectly chain into the parent coroutine.
                WaitForNext(self.parent, self.result)

    def next(self):
        if not self.running:
            return False

        if next(self.routine, False):
            self.started = True
            return True

        self.running = False

        if self.started and self.parent:
            self.parent.result = self.result
            self.parent.next()

        return False

class CoroutineHandler(object):
    def __init__(self, routine):
        self.routine = routine

    def __call__(self, e):
        Coroutine.start1(self.routine, e)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from threading import Lock

class AtomicInt(object):
    def __init__(self, value=0):
        self._value = value
        self._lock = Lock()

    def decrement(self):
        with self._lock:
            self._value -= 1
            return self._value

    def increment(self):
        with self._lock:
            self._value += 1
            return self._value

    @property
    def value(self):
        with self._lock:
            return self._value

    @value.setter
    def value(self, val):
        with self._lock:
            self._value = val



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import inspect
import sys
import types

from x2py.event import Event
from x2py.util.trace import Trace

class EventFactory(object):
    """Manages a map of retrievable event type identifiers and their factory
        methods."""

    class _EventFactory(object):
        def __init__(self):
            self._map = {}

        def create(self, type_id):
            factory_method = self._map.get(type_id)
            if (factory_method is None):
                return None
            return factory_method()

        def register(self, type_id, factory_method):
            if type_id == 0:  # ignore the root event
                return
            existing = self._map.get(type_id)
            if existing:
                if existing != factory_method:
                    raise ValueError("event type id {} conflicted".format(type_id))
                return
            self._map[type_id] = factory_method

        def register_type(self, t):
            EventFactory.register(t.tag.type_id, t)

        def register_module(self, module, base_class=Event):
            predicate = lambda t: inspect.isclass(t) and issubclass(t, base_class)
            members = inspect.getmembers(module, predicate)
            for name, t in members:
                EventFactory.register_type(t)

        def register_package(self, module, base_class=Event):
            EventFactory.register_module(module, base_class)
            for name in dir(module):
                attr = getattr(module, name)
                if type(attr) == types.ModuleType:
                    EventFactory.register_package(attr, base_class)

    instance = _EventFactory()

    def __init__(self):
        raise AssertionError()

    @staticmethod
    def new():
        return EventFactory._EventFactory()

    @staticmethod
    def create(type_id):
        return EventFactory.instance.create(type_id)

    @staticmethod
    def register(type_id, factory_method):
        EventFactory.instance.register(type_id, factory_method)

    @staticmethod
    def register_type(t):
        EventFactory.instance.register_type(t)

    @staticmethod
    def register_module(module, base_class=Event):
        EventFactory.instance.register_module(module, base_class)

    @staticmethod
    def register_package(module, base_class=Event):
        EventFactory.instance.register_package(module, base_class)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import sys
import traceback

from threading import local

from x2py.binding import Binding
from x2py.builtin_events import *
from x2py.case import Case, CaseStack
from x2py.util.trace import Trace

def _init():
    result = local()
    result.current = None
    return result

class Flow(object):
    """Represents a logically independent execution flow."""
    thread_local = _init()

    def __init__(self, name=None):
        self.name = name if name is not None else type(self).__name__
        self.cases = CaseStack()
        self.binding = Binding()

    @staticmethod
    def bind(event, handler):
        return Flow.thread_local.current.subscribe(event, handler)

    @staticmethod
    def unbind(event, handler):
        Flow.thread_local.current.unsubscribe(event, handler)

    @staticmethod
    def unbind_with(token):
        Flow.thread_local.current.unsubscribe(token[0], token[1])

    def attach(self):
        from x2py.hub import Hub
        Hub.instance.attach(self)
        return self

    def detach(self):
        from x2py.hub import Hub
        Hub.instance.detach(self)

        return self

    def add(self, case):
        if case is None or not isinstance(case, Case):
            raise TypeError()
        if self.cases.add(case):
            Trace.debug("flow {}: added case {}", self.name, type(case).__name__)
        return self

    def remove(self, case):
        if case is None or not isinstance(case, Case):
            raise TypeError()
        if self.cases.remove(case):
            Trace.debug("flow {}: removed case {}", self.name, type(case).__name__)
        return self

    def dispatch(self, event):
        event_proxy = Flow.thread_local.event_proxy
        handler_chain = Flow.thread_local.handler_chain

        if len(handler_chain) != 0:
            del handler_chain[:]

        self.binding.build_handler_chain(event, event_proxy, handler_chain)

        for handler in handler_chain:
            try:
                handler(event)
            except Exception as ex:
                Trace.error("flow: dispatch {}".format(ex))
                traceback.print_exc(file=sys.stderr)

        del handler_chain[:]

    def feed(self, event):
        raise NotImplementedError()

    def on_flow_start(self, event):
        self.on_start()
        self.cases.on_start()

    def on_flow_stop(self, event):
        self.cases.on_stop()
        self.on_stop()

    def on_start(self):
        pass

    def on_stop(self):
        pass

    def _setup(self):
        """Called internally when this flow starts up."""

        self.subscribe(FlowStart(), self.on_flow_start)
        self.subscribe(FlowStop(), self.on_flow_stop)

        self.setup()

    def setup(self):
        """Overridden by subclasses to build a flow startup handler chain."""
        pass

    def start(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    def _teardown(self):
        """Called internally when this flow shuts down."""
        self.teardown()

        self.subscribe(FlowStop(), self.on_flow_stop)
        self.subscribe(FlowStart(), self.on_flow_start)

    def teardown(self):
        """Overridden by subclasses to build a flow shutdown handler chain."""
        pass

    def subscribe(self, event, handler):
        return self.binding.bind(event, handler)

    def _unsubscribe(self, event, handler):
        self.binding._unbind(event, handler)

    def unsubscribe(self, event, handler):
        self.binding.unbind(event, handler)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import sys

if sys.version_info.major >= 3:
    from x2py.links.asyncio import *
else:
    from x2py.links.asyncore import *



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.flows.multi_thread_flow import MultiThreadFlow
from x2py.flows.single_thread_flow import SingleThreadFlow
from x2py.flows.threadless_flow import ThreadlessFlow
from x2py.flows.time_flow import TimeFlow



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import codecs

from x2py.buffer_transform import BufferTransform
from x2py.util.trace import Trace

# pycrypto
from Crypto import Random
from Crypto.Cipher import AES, PKCS1_v1_5
from Crypto.PublicKey import RSA

class BlockCipher(BufferTransform):
    """A simple example of BufferTransform that performs block encryption and
        decryption based on the keys exchanged by an asymmetric algorithm."""

    class SettingsBuilder(object):
        def __init__(self, block_size, key_size, rsa_key_size, \
                my_private_key, peer_public_key):
            self.block_size = block_size
            self.key_size = key_size
            self.rsa_key_size = rsa_key_size
            self.my_private_key = my_private_key
            self.peer_public_key = peer_public_key

        def build(self):
            settings = BlockCipher.Settings(self)
            settings.my_rsa = BlockCipher.SettingsBuilder._xml2rsa(self.my_private_key)
            settings.peer_rsa = BlockCipher.SettingsBuilder._xml2rsa(self.peer_public_key)
            return settings

        @staticmethod
        def _xml2rsa(xml):
            import base64
            import xml.etree.ElementTree as etree
            has_private = False
            root = etree.fromstring(xml.replace('\n', ''))
            if root.tag == 'RSAKeyValue':
                for node in root:
                    if node.tag == 'Modulus':
                        n = int(codecs.encode(base64.b64decode(node.text), 'hex'), 16)
                    elif node.tag == 'Exponent':
                        e = int(codecs.encode(base64.b64decode(node.text), 'hex'), 16)
                    elif node.tag == 'P':
                        p = int(codecs.encode(base64.b64decode(node.text), 'hex'), 16)
                    elif node.tag == 'Q':
                        q = int(codecs.encode(base64.b64decode(node.text), 'hex'), 16)
                    elif node.tag == 'DP':
                        pass
                    elif node.tag == 'DQ':
                        pass
                    elif node.tag == 'InverseQ':
                        pass
                    elif node.tag == 'D':
                        d = int(codecs.encode(base64.b64decode(node.text), 'hex'), 16)
                        has_private = True
                if has_private:
                    return RSA.construct((n, e, d, p, q))
                else:
                    return RSA.construct((n, e))
            else:
                raise ValueError()

    class Settings(SettingsBuilder):
        def __init__(self):
            raise AssertionError()

        def __init__(self, builder):
            super(BlockCipher.Settings, self).__init__(
                builder.block_size,
                builder.key_size,
                builder.rsa_key_size,
                builder.my_private_key,
                builder.peer_public_key
            )
            self.my_rsa = None
            self.peer_rsa = None

    def __init__(self, settings=None):
        if settings is not None:
            self.settings = settings
        else:
            # Default settings
            self.settings = BlockCipher.SettingsBuilder(
                block_size = 128,
                key_size = 256,
                rsa_key_size = 1024,
                # In a real-world client/server production, each peer should use
                # a different RSA key pair.
                my_private_key = '''
<RSAKeyValue><Modulus>xtU+mTT9tOES5vLZeSAEvuWaa+FX4jUtH5iVFGSULCaBR6TtQ2TYUz1Jnt
rUhA26OQBIcVzlMyarM8XVhZqk5RJDP64VFz3m+VMmghAgJLUPKDORmIPlc18FuaTsZjxoIwfuVojrDH
/12BoEHHmwb3CVq6dHGsxRLUKG0DYBWQk=</Modulus><Exponent>AQAB</Exponent><P>+3iHfNfD
ARBhnHQ33OyJudsOWkFPwqOG575nkCntjW8RhepXaKPNRqmEu/cYN/Fr/nCmxxgW8Fp5HEI+gI7xZw==
</P><Q>ymoD2gsEj0ksiph+UbkT3Amwx/SHOaRWTwWysL8xKicD0afqnGpHkUnoAUnEQFAnuDIB5D+rb
+6ulwsS6xCsDw==</Q><DP>2d2brJqV1PcnSlAaEepQjFfvwFwzSRM6Ds8UlH7u04k1qkrT/dFkSGMXn
229asJb6O4aYAVL4mLP6J6v3dt54w==</DP><DQ>kLRhtIuT4uupEBwckkgBzpiO7SP/WFIH8c5dBMZq
W3ww2r10mAXSzCdN2T3nMyMagjAd8hMieI7l+c1M5QeyOQ==</DQ><InverseQ>M+sgtHA0blhMUBdGG
IYboxSEvPwPxoX5ORwgL/Zl3TOgxN1oM9i5EkmwKFcazAHKfL5eArtlmfELOcqPMFiyzQ==</Inverse
Q><D>CBEw2AB5ZrRXEv25axusdZ5VNJlQ+oGT0htbuRcXl+78Ac8kPT7DNCVhbkuMocr4ykVDqy3MstW
XzqLxNdl/ZSV9KvP6u5bcDQQeC9KbKQ5PpzGoGmMJNsVtXC0voOA3sYx9P+vVtEqhxn9eAKPOPqX9wRo
9rMW9UZRtDcLiUj0=</D></RSAKeyValue>
''',
                peer_public_key = '''
<RSAKeyValue><Modulus>xtU+mTT9tOES5vLZeSAEvuWaa+FX4jUtH5iVFGSULCaBR6TtQ2TYUz1Jnt
rUhA26OQBIcVzlMyarM8XVhZqk5RJDP64VFz3m+VMmghAgJLUPKDORmIPlc18FuaTsZjxoIwfuVojrDH
/12BoEHHmwb3CVq6dHGsxRLUKG0DYBWQk=</Modulus><Exponent>AQAB</Exponent></RSAKeyVal
ue>
'''
            ).build()

        self.encryptor = None
        self.decryptor = None
        self.encryption_key = None
        self.encryption_iv = None
        self.decryption_key = None
        self.decryption_iv = None

    def cleanup(self):
        pass

    def clone(self):
        return BlockCipher(self.settings)

    @property
    def block_size_in_bytes(self):
        return (self.settings.block_size >> 3)

    @property
    def key_size_in_bytes(self):
        return (self.settings.key_size >> 3)

    def handshake_block_length(self):
        return (self.settings.rsa_key_size >> 3)

    def init_handshake(self):
        challenge_length = self.key_size_in_bytes + self.block_size_in_bytes
        challenge = Random.new().read(challenge_length)
        Trace.trace("challenge: {}", repr(challenge))

        n = self.key_size_in_bytes
        key = challenge[:n]
        iv = challenge[n:n + self.block_size_in_bytes]
        self.encryptor = AES.new(key, AES.MODE_CBC, iv)
        self.encryption_key = key
        self.encryption_iv = iv

        encrypted = self._rsa_encrypt(challenge)
        Trace.trace("encrypted challenge: {}", repr(encrypted))
        return encrypted

    def handshake(self, challenge):
        decrypted = self._rsa_decrypt(challenge)

        n = self.key_size_in_bytes
        key = decrypted[:n]
        iv = decrypted[n:n + self.block_size_in_bytes]
        self.decryptor = AES.new(key, AES.MODE_CBC, iv)
        self.decryption_key = key
        self.decryption_iv = iv

        encrypted = self._rsa_encrypt(decrypted)
        return encrypted

    def fini_handshake(self, response):
        expected = self.encryption_key + self.encryption_iv

        decrypted = self._rsa_decrypt(response)

        return (decrypted == expected)

    def transform(self, buffer):
        buffer = self._pad_pkcs7(buffer)
        return self.encryptor.encrypt(buffer)

    def inverse_transform(self, buffer):
        buffer = self.decryptor.decrypt(bytes(buffer))
        buffer = self._unpad_pkcs7(buffer)
        return buffer

    def _pad_pkcs7(self, buffer):
        length = len(buffer)
        val = self.block_size_in_bytes - (length % self.block_size_in_bytes)
        return bytes(buffer + bytearray([val] * val))

    def _unpad_pkcs7(self, buffer):
        val = buffer[-1]
        if (isinstance(val, str)):
            val = ord(val)
        if val > self.block_size_in_bytes:
            raise ValueError("Invalid PKCS#7 padding")
        length = len(buffer) - val
        return bytes(buffer[:length])

    def _rsa_decrypt(self, ciphertext):
        cipher = PKCS1_v1_5.new(self.settings.my_rsa)
        return cipher.decrypt(ciphertext, None)

    def _rsa_encrypt(self, plaintext):
        cipher = PKCS1_v1_5.new(self.settings.peer_rsa)
        return cipher.encrypt(plaintext)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from collections import deque
from threading import Condition, Lock

from x2py.event import Event

class EventQueue(object):
    """The blocking event queue implementation."""

    def __init__(self):
        self._deque = deque()
        self._closed = False
        self._lock = Lock()
        self._cond = Condition(self._lock)

    def close(self, final_event=None):
        with self._cond:
            if final_event is not None:
                self._deque.append(final_event)
            self._closed = True
            self._cond.notify_all()

    def dequeue(self):
        with self._cond:
            while len(self._deque) == 0:
                if self._closed:
                    return None
                self._cond.wait()
            return self._deque.popleft()

    def enqueue(self, event):
        with self._cond:
            if self._closed:
                return
            self._deque.append(event)
            if len(self._deque) == 1:
                self._cond.notify()

    def try_dequeue(self):
        with self._lock:
            if self._closed:
                return None
            return self._deque.popleft() if len(self._deque) != 0 else None



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from __future__ import print_function

import keyword
import os
import re

from xpiler.formatter import *
from xpiler.types_ import Types

EXTENSION = ".py"
TAB = "    "

def _underscore(s):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1)

def to_snake_case(s):
    return _underscore(s).lower()

def to_SCREAMING_SNAKE_CASE(s):
    return _underscore(s).upper()

def _init_native_types():
    result = {}
    result['bool'] = 'bool'
    result['byte'] = 'int'
    result['int8'] = 'int'
    result['int16'] = 'int'
    result['int32'] = 'int'
    result['int64'] = 'int'
    result['float32'] = 'float'
    result['float64'] = 'float'
    result['string'] = 'str'
    result['datetime'] = 'datetime'
    result['bytes'] = 'bytes'
    result['list'] = 'list'
    result['map'] = 'dict'
    return result

def _init_default_values():
    result = {}
    result['bool'] = 'False'
    result['byte'] = '0'
    result['int8'] = '0'
    result['int16'] = '0'
    result['int32'] = '0'
    result['int64'] = '0'
    result['float32'] = '0.0'
    result['float64'] = '0.0'
    result['string'] = ''
    result['datetime'] = 'datetime(1970, 1, 1)'
    return result

class Python3Formatter(Formatter):
    native_types = _init_native_types()
    default_values = _init_default_values()

    def desc(self):
        return "Python 3"

    def format(self, unit, out_dir):
        basename = to_snake_case(unit.basename)
        target = os.path.join(out_dir, basename + EXTENSION)
        context = Python3FormatterContext(unit, target)
        print(target)
        with open(target, 'w') as f:
            context.out = f
            self._format_head(context)
            self._format_body(context)
            f.flush()

    def is_up_to_date(self, path, out_dir):
        basename = os.path.splitext(os.path.basename(path))[0]
        target = os.path.join(out_dir, basename + EXTENSION)
        return (os.path.exists(target) and
            os.path.getmtime(target) >= os.path.getmtime(path))

    def _format_head(self, context):
        o = context.out
        o.write("# auto-generated by x2py xpiler\n")
        o.write("\n")
        if not context.unit.is_builtin:
            o.write("from x2py.cell import MetaProperty, Cell\n")
            o.write("from x2py.event import Event\n")
            o.write("\n")

        for reference in context.unit.references:
            reference.format(context)
        if len(context.unit.references) != 0:
            o.write("\n")

    def _format_body(self, context):
        o = context.out

        leading = True
        for index, definition in enumerate(context.unit.definitions):
            if index:
                o.write("\n")
            definition.format(context)

class Python3FormatterContext(FormatterContext):
    def __init__(self, unit, target):
        self.unit = unit
        self.target = target
        self.out = None

    def format_cell(self, definition):
        tag_type = 'Event' if definition.is_event else 'Cell'
        #if not self.unit.is_builtin:
        #    tag_type = 'x2py.' + tag_type

        definition.base_class = definition.base
        if definition.base_class is None or len(definition.base_class) == 0:
            definition.base = ''
            definition.base_class = tag_type

        self._preprocess_properties(definition)

        base_tag = definition.base_class
        if len(definition.base) == 0:
            base_tag = definition.base_class + '.tag' if definition.is_event else 'None'
        else:
            base_tag += '.tag'

        tag_initializer_name = '_init_' + to_snake_case(definition.name) + '_tag'
        self._out(0, "def {}():\n".format(tag_initializer_name))
        self._out(1, "props = []\n")
        for prop in definition.properties:
            self._out(1, "props.append({})\n".format(prop.typespec.metaprop(prop.native_name)))
        self._out(1, "return {}.Tag({}, '{}', props".format(tag_type, base_tag, definition.name))
        if definition.is_event:
            self.out.write(",\n")
            if '.' in definition.id:
                tokens = definition.id.split('.')
                tokens[-1] = to_SCREAMING_SNAKE_CASE(tokens[-1])
                definition.id = '.'.join(tokens)
            self._out(2, definition.id)
        self.out.write(")\n\n")

        self._out(0, "class {0}({1}):\n".format(definition.name, definition.base_class))

        self._out(1, "tag = {}()\n\n".format(tag_initializer_name))

        self._format_constructor(definition)
        self._format_properties(definition)
        self._format_methods(definition)

    def format_consts(self, definition):
        if definition.type == 'string':
            for constant in definition.constants:
                constant.value = '"' + constant.value + '"'

        self._out(0, "class {}(object):\n".format(definition.name))
        if definition.constants:
            for constant in definition.constants:
                self._out(1, "{0} = {1}\n".format(
                    to_SCREAMING_SNAKE_CASE(constant.name), constant.value))
        else:
            self._out(1, "pass\n")

    def format_reference(self, reference):
        tokens = reference.target.split('/')
        tokens[-1] = to_snake_case(tokens[-1])
        target = '.'.join(tokens)
        self._out(0, "from {} import *\n".format(target))

    def _preprocess_properties(self, definition):
        index = 0
        for prop in definition.properties:
            prop.index = index
            index += 1

            prop.native_name = to_snake_case(prop.name)
            prop.var_name = '_' + prop.native_name
            if keyword.iskeyword(prop.native_name):
                prop.native_name += '_'

            if Types.is_primitive(prop.typespec.typestr):
                if prop.default_value is None or len(prop.default_value) == 0:
                    prop.default_value = Python3Formatter.default_values[prop.typespec.typestr]
                if prop.typespec.typestr == 'string':
                    prop.default_value = '"' + prop.default_value + '"'
            else:
                prop.default_value = 'None'

            prop.native_type = self._format_typespec(prop.typespec)

    def _format_typespec(self, typespec):
        typestr = typespec.typestr
        if not Types.is_builtin(typestr):
            return typestr  # custom type
        if Types.is_primitive(typestr):
            return Python3Formatter.native_types[typestr]
        else:
            return self._format_collective_type(typespec)

    def _format_collective_type(self, typespec):
        tokens = [ typespec.typestr ]
        if typespec.details is not None:
            tokens.append('(')
            for index, detail in enumerate(typespec.details):
                if index:
                    tokens.append(', ')
                tokens.append(self._format_typespec(detail))
            tokens.append(')')
        return ''.join(tokens)

    def _format_constructor(self, definition):
        self._out(1, "def __init__(self, length=0):\n")
        self._out(2, "super({0}, self).__init__(len({0}.tag.props) + length)\n".format(definition.name))
        self._out(2, "base = {0}.tag.offset\n".format(definition.name))

        if definition.has_properties():
            for index, prop in enumerate(definition.properties):
                self._out(2, "self.values[base + {0}] = {1}\n".format(index, prop.default_value))
            self.out.write("\n")
        else:
            self._out(2, "pass\n")

    def _format_properties(self, definition):
        for index, prop in enumerate(definition.properties):
            if index:
                self.out.write("\n")

            self._out(1, "@property\n")
            self._out(1, "def {}(self):\n".format(prop.native_name))
            self._out(2, "return self.values[{}.tag.offset + {}]\n".format(definition.name, index))
            #self.out.write("\n")
            self._out(1, "@{}.setter\n".format(prop.native_name))
            self._out(1, "def {}(self, value):\n".format(prop.native_name))
            self._out(2, "self._set_property({}.tag.offset + {}, value,\n".format(definition.name, index))
            self._out(3, "{}.tag.props[{}].type_index)\n".format(definition.name, index))

    def _format_methods(self, definition):
        self._format_type(definition)

    def _format_type(self, definition):
        self.out.write("\n")
        self._out(1, "def type_id(self):\n")
        self._out(2, "return {}.tag.type_id\n".format(definition.name))
        self.out.write("\n")
        self._out(1, "def type_tag(self):\n")
        self._out(2, "return {}.tag\n".format(definition.name))

    def _out(self, indentation, s):
        self._indent(indentation)
        self.out.write(s)

    def _indent(self, indentation):
        for _ in range(indentation):
            self.out.write(TAB)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.event_factory import EventFactory
from x2py.hub import Hub
from x2py.link import Link
from x2py.util.ranged_int_pool import RangedIntPool
from x2py.util.rwlock import ReadLock, WriteLock, ReadWriteLock
from x2py.util.trace import Trace

from x2py.links.link_events import *

def _static_init():
    EventFactory.register(LinkEventType.HANDSHAKE_REQ, HandshakeReq)
    EventFactory.register(LinkEventType.HANDSHAKE_RESP, HandshakeResp)
    EventFactory.register(LinkEventType.HANDSHAKE_ACK, HandshakeAck)

    return RangedIntPool(1, 65536, True)

class SessionBasedLink(Link):
    handle_pool = _static_init()

    def __init__(self, name):
        super(SessionBasedLink, self).__init__(name)
        self.rwlock = ReadWriteLock()
        self.channel_strategy = None
        self.heartbeat_strategy = None

    @property
    def has_channel_strategy(self):
        return (self.channel_strategy is not None)

    @property
    def has_heartbeat_strategy(self):
        return (self.heartbeat_strategy is not None)

    def init_session(self, session):
        if self.has_channel_strategy:
            self.channel_strategy.before_session_setup(session)
        if self.has_heartbeat_strategy:
            self.heartbeat_strategy.before_session_setup(session)

        if self.has_channel_strategy:
            self.channel_strategy.init_handshake(session)
        else:
            self.on_connect(True, session)


    def on_connect(self, result, context):
        Trace.info("{} connected {} {}", self.name, result, context)

        if result:
            handle = SessionBasedLink.handle_pool.acquire()
            context.handle = handle

        self._on_connect(result, context)

        LinkSessionConnected().setattrs(
            link_name = self.name,
            result = result,
            context = context
        ).post()

    def on_disconnect(self, handle, context):
        Trace.info("{} disconnected {} {}", self.name, handle, context)

        self._on_disconnect(handle, context)

        if handle != 0:
            SessionBasedLink.handle_pool.release(handle)

        LinkSessionDisconnected().setattrs(
            link_name = self.name,
            handle = handle,
            context = context
        ).post()

    def _on_connect(self, result, context):
        pass

    def _on_disconnect(self, handle, context):
        pass

    def _setup(self):
        super(SessionBasedLink, self)._setup()

        self.bind(LinkSessionConnected().setattrs(link_name = self.name),
            self.on_link_session_connected)
        self.bind(LinkSessionDisconnected().setattrs(link_name = self.name),
            self.on_link_session_disconnected)

        if self.has_channel_strategy:
            self.channel_strategy.link = self
            self.channel_strategy.setup()

        if self.has_heartbeat_strategy:
            self.heartbeat_strategy.link = self
            self.heartbeat_strategy.setup()

            self.bind(Hub.heartbeat_event, self.on_heartbeat_event)

    def _teardown(self):
        if self.has_heartbeat_strategy:
            self.heartbeat_strategy.teardown()
            self.heartbeat_strategy = None
        if self.has_channel_strategy:
            self.channel_strategy.teardown()
            self.channel_strategy = None

        self.unbind(LinkSessionConnected().setattrs(link_name = self.name),
            self.on_link_session_connected)
        self.unbind(LinkSessionDisconnected().setattrs(link_name = self.name),
            self.on_link_session_disconnected)

        super(SessionBasedLink, self)._teardown()

    def on_link_session_connected(self, event):
        self.on_session_connected(event.result, event.context)

    def on_link_session_disconnected(self, event):
        self.on_session_disconnected(event.handle, event.context)

    def on_session_connected(self, result, context):
        pass

    def on_session_disconnected(slef, handle, context):
        pass

    def on_heartbeat_event(self, event):
        self.heartbeat_strategy.on_heartbeat()



import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

from x2py import __version__

class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def run_tests(self):
        import shlex
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)

setup(
    name='x2py',
    version=__version__,
    description='Python port of x2',
    url='https://github.com/jaykang920/x2py',
    author='Jae-jun Kang',
    author_email='jaykang920@gmail.com',
    license='MIT',
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
    packages=[
        'x2py',
        'x2py.flows',
        'x2py.links',
        'x2py.links.asyncio',
        'x2py.links.asyncore',
        'x2py.transforms',
        'x2py.util',
        'x2py.yields',
        'xpiler',
    ],
    scripts=['scripts/x2py.xpiler.py']
)


# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import datetime
import struct

class Deserializer(object):
    """Binary wire format deserializer."""

    def read_bool(self, metaprop):
        self.check_length(1)
        b = self.buffer[self.pos]
        if isinstance(b, str):
            b = ord(b)
        self.pos += 1
        return True if b != 0 else False

    def read_byte(self, metaprop):
        self.check_length(1)
        b = self.buffer[self.pos]
        if isinstance(b, str):
            b = ord(b)
        self.pos += 1
        return b

    def read_int8(self, metaprop):
        self.check_length(1)
        offset = self.pos
        self.pos += 1
        return struct.unpack_from('b', self.buffer, offset)[0]

    def read_int16(self, metaprop):
        self.check_length(2)
        offset = self.pos
        self.pos += 2
        return struct.unpack_from('!h', self.buffer, offset)[0]

    def read_int32(self, metaprop):
        value, _ = self.read_variable32()
        value = (value >> 1) ^ -(value & 1)
        if value < -(2**31) or (2**31 - 1) < value:
            raise ValueError()
        return value

    def read_int64(self, metaprop):
        value, _ = self.read_variable64()
        value = (value >> 1) ^ -(value & 1)
        if value < -(2**63) or (2**63 - 1) < value:
            raise ValueError()
        return value

    def read_float32(self, metaprop):
        self.check_length(4)
        offset = self.pos
        self.pos += 4
        return struct.unpack_from('!f', self.buffer, offset)[0]

    def read_float64(self, metaprop):
        self.check_length(8)
        offset = self.pos
        self.pos += 8
        return struct.unpack_from('!d', self.buffer, offset)[0]

    def read_string(self, metaprop):
        length, _ = self.read_nonnegative()
        if length == 0:
            return ''
        temp = self.buffer[self.pos:self.pos + length]
        self.pos += length
        return temp.decode('utf-8')

    def read_datetime(self, metaprop):
        self.check_length(8)
        offset = self.pos
        self.pos += 8
        millisecs = struct.unpack_from('!q', self.buffer, offset)[0]
        unix_epoch = datetime.datetime(1970, 1, 1)
        return unix_epoch + datetime.timedelta(milliseconds=millisecs)

    def read_bytes(self, metaprop):
        length, _ = self.read_nonnegative()
        if length == 0:
            return None
        value = self.buffer[self.pos:self.pos + length]
        self.pos += length
        return value

    def read_cell(self, metaprop):
        from x2py.event import Event
        length, _ = self.read_nonnegative()
        if length == 0:
            return None
        temp = self.buffer[self.pos:self.pos + length]
        self.pos += length
        is_event = issubclass(metaprop.runtime_type, Event)
        if is_event:
            type_id = self.read_int32(None)
            value = EventFactory.create(type_id)
        else:
            value = metaprop.runtime_type()
        if value is not None:
            value.deserialize(Deserializer(temp))
        return value

    def read_list(self, metaprop):
        result = []
        length, _ = self.read_nonnegative()
        if length == 0:
            return result
        for i in range(length):
            result.append(self.read(metaprop.details[0]))
        return result

    def read_map(self, metaprop):
        result = {}
        length, _ = self.read_nonnegative()
        if length == 0:
            return result
        for i in range(length):
            key = self.read(metaprop.details[0])
            value = self.read(metaprop.details[1])
            result[key] = value
        return result

    # Reader function table
    readers = [
        None,
        read_bool,
        read_byte,
        read_int8,
        read_int16,
        read_int32,
        read_int64,
        read_float32,
        read_float64,
        read_string,
        read_datetime,
        read_bytes,
        read_cell,
        read_list,
        read_map,
        None  # none for object type
    ]

    def __init__(self, buffer=None):
        self.buffer = buffer
        if self.buffer is None:
            self.buffer = bytearray()
        self.pos = 0

    def read(self, metaprop):
        reader = Deserializer.readers[metaprop.type_index]
        return reader(self, metaprop)

    def read_nonnegative(self):
        value, num_bytes = Deserializer.read_variable32(self)
        if value < 0:
            raise ValueError()
        return value, num_bytes

    def read_variable32(self):
        return self._read_variable(5)

    def read_variable64(self):
        return self._read_variable(10)

    def _read_variable(self, max_bytes):
        value = 0
        i = shift = 0
        while i < max_bytes:
            self.check_length(1)
            b = self.buffer[self.pos]
            if isinstance(b, str):
                b = ord(b)
            self.pos += 1
            value = value | ((b & 0x7f) << shift)
            if (b & 0x80) == 0:
                break
            i += 1
            shift += 7
        return value, min(i + 1, max_bytes)

    def check_length(self, num_bytes):
        if (self.pos + num_bytes) > len(self.buffer):
            raise EOFError()



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import datetime
import sys

from x2py.fingerprint import Fingerprint
from x2py.serializer import Serializer
from x2py.util.hash import Hash

class MetaProperty(object):
    """Represents runtime traits of a cell property."""

    BOOL = 1
    BYTE = 2
    INT8 = 3
    INT16 = 4
    INT32 = 5
    INT64 = 6
    FLOAT32 = 7
    FLOAT64 = 8
    STRING = 9
    DATETIME = 10
    BYTES = 11
    CELL = 12
    LIST = 13
    MAP = 14
    OBJECT = 15  # not serializable

    def __init__(self, name, type_index, runtime_type=None, details=None):
        self.name = name
        self.type_index = type_index
        self.runtime_type = runtime_type
        self.details = details  # list of child MetaProperty objects

class Cell(object):
    """Common base class for all custom types."""

    class Tag(object):
        def __init__(self, base, type_name, props):
            self.base = base
            self.type_name = type_name
            self.props = props
            self.offset = 0
            if base is not None:
                self.offset = base.offset + len(base.props)

    tag = Tag(None, 'Cell', [])

    def __init__(self, length):
        self.fingerprint = Fingerprint(length)
        self.values = [None] * length  # property values

    def desc(self):
        prop_descs = []
        tag = self.type_tag()
        self._desc(tag, prop_descs)
        result = ', '.join(prop_descs)
        result = "{} {{ {} }}".format(tag.type_name, result)
        return result

    def _desc(self, tag, prop_descs):
        if tag.base is not None:
            self._desc(tag.base, prop_descs)
        if len(tag.props) == 0:
            return
        for index, prop in enumerate(tag.props):
            if prop.name.startswith('_'):
                continue
            value = self.values[tag.offset + index]
            if prop.type_index == MetaProperty.STRING:
                value = "'{}'".format(value.replace("'", "''"))
            elif prop.type_index == MetaProperty.BYTES:
                value = repr(value)
            prop_descs.append('{}:{}'.format(prop.name, value))

    def deserialize(self, deserializer):
        self.fingerprint.deserialize(deserializer)
        self._deserialize(self.type_tag(), deserializer)

    def _deserialize(self, tag, deserializer):
        if tag.base is not None:
            self._deserialize(tag.base, deserializer)
        if len(tag.props) == 0:
            return
        base = tag.offset
        for index, prop in enumerate(tag.props):
            if prop.name.startswith('_'):
                if prop.name == '_Handle':
                    continue
            if self.fingerprint.get(base + index):
                self.values[base + index] = deserializer.read(prop)

    def get_length(self, target_type=None):
        result = self.fingerprint.get_length()
        length, _ = self._get_length(self.type_tag(), target_type, True)
        result += length
        return result

    def _get_length(self, tag, target_type, flag):
        result = 0
        if tag.base is not None:
            length, flag = self._get_length(tag.base, target_type, flag)
            result += length
            if not flag:
                return result, flag
        if len(tag.props) == 0:
            return result, flag
        base = tag.offset
        for index, prop in enumerate(tag.props):
            if prop.name.startswith('_'):
                if prop.name == '_Handle':
                    continue
            if self.fingerprint.get(base + index):
                result += Serializer.get_length(prop, self.values[base + index])
        if (target_type is not None) and (target_type.__name__ == tag.type_name):
            flag = False
        return result, flag

    def type_tag(self):
        return Cell.tag

    def equals(self, other):
        if self is other:
            return True
        if type(self) != type(other):
            return False
        return self._equals(self.type_tag(), other)

    def _equals(self, tag, other):
        if tag.base is not None:
            if not self._equals(tag.base, other):
                return False
        if len(tag.props) == 0:
            return True
        base = tag.offset
        for index, prop in enumerate(tag.props):
            if self.values[base + index] != other.values[base + index]:
                return False
        return True

    def equivalent(self, other):
        if self is other:
            return True
        if not isinstance(self, type(other)):
            return False
        return self._equivalent(self.type_tag(), other)

    def _equivalent(self, tag, other):
        if tag.base is not None:
            if not self._equivalent(tag.base, other):
                return False
        if len(tag.props) == 0:
            return True
        base = tag.offset
        for index, prop in enumerate(tag.props):
            if other.fingerprint.get(base + index):
                if self.values[base + index] != other.values[base + index]:
                    return False
        return True

    def hash_code(self, fingerprint):
        h = Hash()
        self._hash_code(self.type_tag(), h, fingerprint)
        return h.code

    def _hash_code(self, tag, h, fingerprint):
        if tag.base is not None:
            self._hash_code(tag.base, h, fingerprint)
        if len(tag.props) == 0:
            return
        base = tag.offset
        for index, prop in enumerate(tag.props):
            if fingerprint.get(base + index):
                h.update(base + index)                     # property index
                h.update(hash(self.values[base + index]))  # property value

    def serialize(self, serializer, target_type=None):
        self.fingerprint.serialize(serializer)
        self._serialize(self.type_tag(), serializer, target_type, True)

    def _serialize(self, tag, serializer, target_type, flag):
        if tag.base is not None:
            flag = self._serialize(tag.base, serializer, target_type, flag)
            if not flag:
                return flag
        if len(tag.props) == 0:
            return flag
        base = tag.offset
        for index, prop in enumerate(tag.props):
            if prop.name.startswith('_'):
                if prop.name == '_Handle':
                    continue
            if self.fingerprint.get(base + index):
                serializer.write(prop, self.values[base + index])
        if (target_type is not None) and (target_type.__name__ == tag.type_name):
            flag = False
        return flag

    def setattrs(self, **kwargs):
        """Sets multiple attributes of this cell object at once."""
        for key, value in kwargs.items():
            setattr(self, key, value)
        return self

    def __eq__(self, other):
        return other.equals(self)

    def __ne__(self, other):
        return not other.equals(self)

    def __hash__(self):
        return self.hash_code(self.fingerprint)

    def __str__(self):
        return self.desc()

    # Property type checkers

    @staticmethod
    def is_bool(value):
        if value == True or value == False:
            return True, value
        else:
            return False, value

    @staticmethod
    def is_byte(value):
        if not isinstance(value, int):
            try:
                value = int(value)
            except:
                return False, value
        if value < 0 or (2**8 - 1) < value:
            return False, value
        return True, value

    @staticmethod
    def is_int8(value):
        if not isinstance(value, int):
            try:
                value = int(value)
            except:
                return False, value
        if value < -(2**7) or (2**7 - 1) < value:
            return False, value
        return True, value

    @staticmethod
    def is_int16(value):
        if not isinstance(value, int):
            try:
                value = int(value)
            except:
                return False, value
        if value < -(2**15) or (2**15 - 1) < value:
            return False, value
        return True, value

    @staticmethod
    def is_int32(value):
        if not isinstance(value, int):
            try:
                value = int(value)
            except:
                return False, value
        if value < -(2**31) or (2**31 - 1) < value:
            return False, value
        return True, value

    @staticmethod
    def is_int64(value):
        if not isinstance(value, int):
            try:
                value = int(value)
            except:
                return False, value
        if value < -(2**63) or (2**63 - 1) < value:
            return False, value
        return True, value

    @staticmethod
    def is_float32(value):
        if not isinstance(value, float):
            try:
                value = float(value)
            except:
                return False, value
        return True, value

    @staticmethod
    def is_float64(value):
        if not isinstance(value, float):
            try:
                value = float(value)
            except:
                return False, value
        return True, value

    @staticmethod
    def is_string(value):
        if value is not None:  # allow None
            if not isinstance(value, str):
                try:
                    value = str(value)
                except:
                    return False, value
        return True, value

    @staticmethod
    def is_datetime(value):
        if not isinstance(value, datetime.datetime):
            try:
                value = datetime.datetime(value)
            except:
                return False, value
        return True, value

    @staticmethod
    def is_bytes(value):
        if value is not None:  # allow None
            if not isinstance(value, bytes):
                try:
                    value = bytes(value)
                except:
                    return False, value
        return True, value

    @staticmethod
    def is_cell(value):
        if value is not None:  # allow None
            if not isinstance(value, Cell):
                return False, value
        return True, value

    @staticmethod
    def is_list(value):
        if value is not None:  # allow None
            if not isinstance(value, list):
                return False, value
        return True, value

    @staticmethod
    def is_map(value):
        if value is not None:  # allow None
            if not isinstance(value, dict):
                return False, value
        return True, value

    @staticmethod
    def is_object(value):
        return True, value

    # Property type checker function table
    checkers = [
        None,
        is_bool,
        is_byte,
        is_int8,
        is_int16,
        is_int32,
        is_int64,
        is_float32,
        is_float64,
        is_string,
        is_datetime,
        is_bytes,
        is_cell,
        is_list,
        is_map,
        is_object
    ]

    def _set_property(self, index, value, type_index):
        checker = Cell.checkers[type_index]
        valid, value = checker.__func__(value)
        if not valid:
            raise ValueError()
        if value is not None:
            self.fingerprint.touch(index)
        self.values[index] = value



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

import weakref
from threading import Lock

from x2py.util.trace import Trace

class EventSink(object):
    """Cleanup helper base class for any event-consuming classes."""

    def __init__(self):
        self._bindings = []
        self._disposed = False
        self._flow = None
        from x2py.flow import Flow
        flow = Flow.thread_local.current
        if flow is not None:
            self._flow = weakref.ref(flow)
        self._lock = Lock()

    @property
    def flow(self):
        return self._flow() if self._flow is not None else None

    @flow.setter
    def flow(self, value):
        if self._disposed or len(self._bindings) != 0:
            raise RuntimeError()
        self._flow = weakref.ref(value) if value is not None else None

    def cleanup(self):
        with self._lock:
            if self._disposed:
                return

            flow = self.flow
            if flow is None:
                return

            try:
                for binding in self._bindings:
                    flow._unsubscribe(binding[0], binding[1])
                del self._bindings[:]

                self.flow = None
            except BaseException as ex:
                Trace.warn("eventsink: cleanup error {}", ex)
            finally:
                self_disposed = True

    def bind(self, event, handler):
        flow = self.flow
        if flow is not None:
            flow.subscribe(event, handler)

    def unbind(self, event, handler):
        flow = self.flow
        if flow is not None:
            flow.unsubscribe(event, handler)

    def _add_binding(self, token):
        with self._lock:
            self._bindings.append(token)

    def _remove_binding(self, token):
        with self._lock:
            self._bindings.remove(token)





# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from threading import RLock, Semaphore

class WriteLock(object):
    def __init__(self, rwlock):
        self.rwlock = rwlock

    def __enter__(self):
        self.rwlock.wlock_acquire()
        return self

    def __exit__(self, type, value, traceback):
        self.rwlock.wlock_release()

class ReadLock(object):
    def __init__(self, rwlock):
        self.rwlock = rwlock
        self.upgraded = False

    def upgrade(self):
        self.rwlock.rlock_release()
        self.rwlock.wlock_acquire()
        self.upgraded = True

    def __enter__(self):
        self.rwlock.rlock_acquire()
        return self

    def __exit__(self, type, value, traceback):
        if self.upgraded:
            self.rwlock.wlock_release()
        else:
            self.rwlock.rlock_release()

class ReadWriteLock(object):
    """Reader-writer lock with preference to writers."""

    def __init__(self):
        self._lock = RLock()
        self._can_read = Semaphore(0)
        self._can_write = Semaphore(0)
        self._active_readers  = 0
        self._active_writers  = 0
        self._waiting_readers = 0
        self._waiting_writers = 0

    def rlock(self):
        return ReadLock(self)

    def rlock_acquire(self):
        with self._lock:
            if self._active_writers == 0 and self._waiting_writers == 0:
                self._active_readers += 1
                self._can_read.release()
            else:
                self._waiting_readers += 1
        self._can_read.acquire()

    def rlock_release(self):
        with self._lock:
            self._active_readers -= 1
            if self._active_readers == 0 and self._waiting_writers != 0:
                self._active_writers  += 1
                self._waiting_writers -= 1
                self._can_write.release()

    def wlock(self):
        return WriteLock(self)

    def wlock_acquire(self):
        with self._lock:
            if self._active_writers == 0 and self._waiting_writers == 0 and self._active_readers == 0:
                self._active_writers += 1
                self._can_write.release()
            else:
                self._waiting_writers += 1
        self._can_write.acquire()

    def wlock_release(self):
        with self._lock:
            self._active_writers -= 1
            if self._waiting_writers != 0:
                self._active_writers  += 1
                self._waiting_writers -= 1
                self._can_write.release()
            elif self._waiting_readers != 0:
                t = self._waiting_readers
                self._waiting_readers = 0
                self._active_readers += t
                while t > 0:
                    self._can_read.release()
                    t -= 1



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from datetime import timedelta

from x2py.builtin_events import TimeoutEvent
from x2py.config import Config
from x2py.flow import Flow
from x2py.flows.time_flow import TimeFlow
from x2py.util.trace import Trace
from x2py.yields.wait_handle_pool import WaitHandlePool

class WaitForEvent(object):
    def __init__(self, coroutine, e, seconds=Config.Coroutine.default_timeout):
        self.coroutine = coroutine
        self.handler_token = Flow.bind(e, self.on_event)
        self.timerout_token = None
        self.timer_token = None
        if seconds > 0:
            timeout_event = TimeoutEvent().setattrs(key=self)
            self.timeout_token = Flow.bind(timeout_event, self.on_timeout)
            self.timer_token = TimeFlow.get().reserve(timeout_event, timedelta(seconds=seconds))

    def on_event(self, e):
        Flow.unbind_with(self.handler_token)

        if self.timer_token:
            TimeFlow.get().cancel(self.timer_token)
            Flow.unbind_with(self.timeout_token)

        wait_handle = self.handler_token[0]._wait_handle
        if wait_handle:
            WaitHandlePool.release(wait_handle)

        coroutine.result = e
        coroutine.next()

    def on_timeout(self, e):
        Flow.unbind_with(self.handler_token)
        Flow.unbind_with(self.timeout_token)

        wait_handle = self.handler_token[0]._wait_handle
        if wait_handle:
            WaitHandlePool.release(wait_handle)

        Trace.error("WaitForEvent timeout for {}".format(self.handler_token[0]))

        self.coroutine.result = None
        self.coroutine.next()

class WaitForResponse(WaitForEvent):
    def __init__(self, coroutine, req, e, seconds=Config.Coroutine.default_timeout):
        wait_handle = WaitHandlePool.acquire()
        req._wait_handle = wait_handle
        e._wait_handle = wait_handle

        super(WaitForResponse, self).__init__(coroutine, e, seconds)

        req.post()



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from x2py.cell import MetaProperty, Cell
from x2py.util.hash import hash_update

class Event(Cell):
    """Common base class for all events."""

    class Tag(Cell.Tag):
        def __init__(self, base, type_name, props, type_id):
            Cell.Tag.__init__(self, base, type_name, props)
            self.type_id = type_id

    tag = Tag(None, 'Event', [
            MetaProperty('_Handle', MetaProperty.INT32),
            MetaProperty('_WaitHandle', MetaProperty.INT32)
        ], 0)

    def __init__(self, length=0):
        super(Event, self).__init__(len(Event.tag.props) + length)
        self._transform = True
        base = Event.tag.offset
        self.values[base + 0] = 0
        self.values[base + 1] = 0

    @property
    def _handle(self):
        return self.values[Event.tag.offset + 0]
    @_handle.setter
    def _handle(self, value):
        self._set_property(Event.tag.offset + 0, value,
            Event.tag.props[0].type_index)

    @property
    def _wait_handle(self):
        return self.values[Event.tag.offset + 1]
    @_wait_handle.setter
    def _wait_handle(self, value):
        self._set_property(Event.tag.offset + 1, value,
            Event.tag.props[1].type_index)

    def desc(self):
        prop_descs = []
        tag = self.type_tag()
        self._desc(tag, prop_descs)
        result = ', '.join(prop_descs)
        result = "{} {} {{ {} }}".format(tag.type_id, tag.type_name, result)
        return result

    def in_response_of(self, request):
        self._handle = request._handle
        return self

    def post(self):
        """Posts up this event to the hub."""
        from x2py.hub import Hub
        Hub.post(self)

    def type_id(self):
        return Event.tag.type_id

    def type_tag(self):
        return Event.tag

    def hash_code_for(self, fingerprint, type_id):
        value = self.hash_code(fingerprint)
        value = hash_update(value, -1)  # delimiter for type id
        value = hash_update(value, type_id)
        return value

    def __hash__(self):
        return self.hash_code_for(self.fingerprint, self.type_id())

class EventProxy(object):
    """Supports dictionary search by equivalence."""

    def __init__(self):
        self.event = None
        self.fingerprint = None
        self.type_id = 0

    def equals(self, other):
        return self.event.equivalent(other)

    def __hash__(self):
        return self.event.hash_code_for(self.fingerprint, self.type_id)



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.
"""Import core names of x2py."""

__version__ = '0.4.2'

from x2py.buffer_transform import BufferTransform
from x2py.builtin_events import *
from x2py.case import Case
from x2py.config import Config
from x2py.coroutine import Coroutine, CoroutineHandler
from x2py.event import Event
from x2py.event_factory import EventFactory
from x2py.event_sink import EventSink
from x2py.flow import Flow
from x2py.hub import Hub
from x2py.link import Link

from x2py.flows import *
from x2py.util import *
from x2py.yields import *



# Copyright (c) 2017, 2018 Jae-jun Kang
# See the file LICENSE for details.

from array import array
from threading import Lock

class RangedIntPool(object):
    def __init__(self, min_value, max_value, advancing=False):
        self._min_value = min_value
        self._max_value = max_value
        self._advancing = advancing
        self._offset = 0
        self._blocks = array('L', [0] * (((max_value - min_value) >> 5) + 1))
        self._lock = Lock()

    def acquire(self):
        with self._lock:
            index = self._offset
            length = self._max_value - self._min_value + 1
            i = 0
            while i < length:
                if (index >= length):
                    index = 0
                if not self._get(index):
                    self._set(index, True)
                    if self._advancing:
                        self._offset = index + 1
                        if (self._offset >= length):
                            self._offset = 0
                    return self._min_value + index
                i += 1
                index += 1
            raise MemoryError()

    def claim(self, value):
        if value < self._min_value or self._max_value < value:
            raise ValueError()
        index = value - self._min_value
        with self._lock:
            if (self._get(index)):
                return False
            self._set(index, True)
        return True

    def release(self, value):
        if value < self._min_value or self._max_value < value:
            raise ValueError()
        index = value - self._min_value
        with self._lock:
            self._set(index, False)

    def _get(self, index):
        return ((self._blocks[index >> 5] & (1 << (index & 0x1f))) != 0)

    def _set(self, index, flag):
        if flag:
            self._blocks[index >> 5] |= (1 << (index & 0x1f))
        else:
            self._blocks[index >> 5] &= ~(1 << (index & 0x1f))
